# -*- coding: utf-8 -*-
r"""Tools for Fitter class
"""
from functools import wraps

import numpy as np


def convert_params(params):
    r"""

    Parameters
    ----------
    params : obj, list
        Will convert a LMFIT Parameters object or pass parameters list into a
        list of parameters.

    Returns
    -------

    """
    if not isinstance(params, (list, np.ndarray)):
        parvals = params.valuesdict()
        parin = [parvals['p{0}'.format(i)] for i in range(len(parvals))]
    else:
        parin = params

    return parin


def residual_wrapper(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        parin = convert_params(args[0])
        argin = (parin,) + args[1:]
        return function(*argin, **kwargs)
    return wrapper



# -*- coding: utf-8 -*-
r"""Chopper class for Time of Flight instrument

"""
import numpy as np

from .exceptions import ChopperError


class Chopper(object):
    r"""Class defining a chopper object for using Time of Flight spectrometer
    resolution calculations.

    Parameters
    ----------
    distance : float
        Distance of the chopper from the source in cm.

    speed : float
        Speed of the chopper in Hz.

    width : float
        width of the beam at the chopper in cm.

    chopper_type : string
        The type of chopper: 'disk' or 'fermi'.

    acceptance : float
        If chopper_type == 'disk', angular acceptance of the chopper in
        degrees, unless `radius` is defined, in which case `acceptance`
        is the size of the opening in the disk in cm. If chopper_type ==
        'fermi', distance between chopper blades in cm.

    counter_rot : bool, optional
        If the disk chopper consists of two counter rotating choppers, set to
        True (Default: False).

    radius : float, optional
        radius of the chopper in cm. If defined, and chopper_type == 'disk',
        then `acceptance` is assumed to be in units of cm.

    depth : float, optional
        The depth of the Fermi Chopper blades, to calculate angular
        acceptance. Required if chopper_type == 'fermi'.

    tau : float, optional
        Custom value of the resolution of the chopper in standard deviation in
        units of microseconds. Used to override the automatic calculation of
        tau.

    Attributes
    ----------
    distance
    speed
    width
    chopper_type
    acceptance
    radius
    depth
    tau

    """
    def __init__(self, distance, speed, width, chopper_type, acceptance, counter_rot=False, radius=None, depth=None, tau=None):
        self.distance = distance
        self.speed = speed
        self.width = width
        self.chopper_type = chopper_type
        self.acceptance = acceptance
        if counter_rot:
            self.counter_rot = 2.0
        else:
            self.counter_rot = 1.0
        if radius is not None:
            self.radius = radius
        if depth is not None:
            self.depth = depth
        if tau is not None:
            self.tau_override = tau


    def __repr__(self):
        args = ', '.join(
            [str(getattr(self, key)) for key in ['distance', 'speed', 'width', 'chopper_type', 'acceptance']])
        kwargs = ', '.join(
            ['{0}={1}'.format(getattr(self, key)) for key in ['depth', 'tau'] if getattr(self, key, None) is not None])
        return "Chopper({0})".format(', '.join([args, kwargs]))

    @property
    def tau(self):
        """Calculate the time resolution of the chopper

        Returns
        -------
        tau : float
            Returns the resolution of the chopper in standard deviation in units of microseconds
        """
        if hasattr(self, 'tau_override'):
            return self.tau_override

        elif self.chopper_type == 'disk' and hasattr(self, 'radius'):
            return self.acceptance / (self.radius * self.speed * self.counter_rot) / np.sqrt(8 * np.log(2))
        elif self.chopper_type == 'disk' and ~hasattr(self, 'radius'):
            return 1e6 / (self.speed * self.acceptance * self.counter_rot) / 360.0
        elif self.chopper_type == 'fermi':
            try:
                return 1e6 / (self.speed * 2.0 * np.arctan(self.acceptance / self.depth)) / 360.
            except AttributeError:
                raise ChopperError("'depth' not specified, and is a required value for a Fermi Chopper.")

        else:
            raise ChopperError("'{0}' is an invalid chopper_type. Choose 'disk' or 'fermi', or specify custom tau \
                                via `tau_override` attribute".format('chopper_type'))



# -*- coding: utf-8 -*-
r"""Define an instrument for resolution calculations

"""
import numpy as np
from scipy.linalg import block_diag as blkdiag

from ..crystal import Sample
from ..energy import Energy
from .analyzer import Analyzer
from .exceptions import ScatteringTriangleError
from .general import GeneralInstrument
from .monochromator import Monochromator
from .plot import PlotInstrument
from .tools import GetTau, _CleanArgs, _Dummy, _modvec, _scalar, _star, _voigt


class TripleAxisInstrument(GeneralInstrument, PlotInstrument):
    u"""An object that represents a Triple Axis Spectrometer (TAS) instrument
    experimental configuration, including a sample.

    Parameters
    ----------
    efixed : float, optional
        Fixed energy, either ei or ef, depending on the instrument
        configuration. Default: 14.7

    sample : obj, optional
        Sample lattice constants, parameters, mosaic, and orientation
        (reciprocal-space orienting vectors). Default: A crystal with
        a,b,c = 6,7,8 and alpha,beta,gamma = 90,90,90 and orientation
        vectors u=[1 0 0] and v=[0 1 0].

    hcol : list(4)
        Horizontal Soller collimations in minutes of arc starting from the
        neutron guide. Default: [40 40 40 40]

    vcol : list(4), optional
        Vertical Soller collimations in minutes of arc starting from the
        neutron guide. Default: [120 120 120 120]

    mono_tau : str or float, optional
        The monochromator reciprocal lattice vector in Å\ :sup:`-1`,
        given either as a float, or as a string for common monochromator types.
        Default: 'PG(002)'

    mono_mosaic : float, optional
        The mosaic of the monochromator in minutes of arc. Default: 25

    ana_tau : str or float, optional
        The analyzer reciprocal lattice vector in Å\ :sup:`-1`,
        given either as a float, or as a string for common analyzer types.
        Default: 'PG(002)'

    ana_mosaic : float, optional
        The mosaic of the monochromator in minutes of arc. Default: 25

    Attributes
    ----------
    method
    moncor
    mono
    ana
    hcol
    vcol
    arms
    efixed
    sample
    orient1
    orient2
    infin
    beam
    detector
    monitor
    Smooth
    guide
    description_string

    Methods
    -------
    calc_resolution
    calc_resolution_in_Q_coords
    calc_projections
    get_angles_and_Q
    get_lattice
    get_resolution_params
    get_resolution
    plot_projections
    plot_ellipsoid
    plot_instrument
    resolution_convolution
    resolution_convolution_SMA
    plot_slice

    """
    def __init__(self, efixed=14.7, sample=None, hcol=None, vcol=None, mono='PG(002)',
                 mono_mosaic=25, ana='PG(002)', ana_mosaic=25, **kwargs):

        if sample is None:
            sample = Sample(6, 7, 8, 90, 90, 90)
            sample.u = [1, 0, 0]
            sample.v = [0, 1, 0]

        if hcol is None:
            hcol = [40, 40, 40, 40]

        if vcol is None:
            vcol = [120, 120, 120, 120]

        self.mono = Monochromator(mono, mono_mosaic)
        self.ana = Analyzer(ana, ana_mosaic)
        self.hcol = np.array(hcol)
        self.vcol = np.array(vcol)
        self.efixed = efixed
        self.sample = sample
        self.orient1 = np.array(sample.u)
        self.orient2 = np.array(sample.v)
        self.detector = _Dummy('Detector')
        self.monitor = _Dummy('Monitor')
        self.guide = _Dummy('Guide')

        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return "Instrument('tas', engine='neutronpy', efixed={0})".format(self.efixed)

    def __eq__(self, right):
        self_parent_keys = sorted(list(self.__dict__.keys()))
        right_parent_keys = sorted(list(right.__dict__.keys()))

        if not np.all(self_parent_keys == right_parent_keys):
            return False

        for key, value in self.__dict__.items():
            right_parent_val = getattr(right, key)
            if not np.all(value == right_parent_val):
                print(value, right_parent_val)
                return False

        return True

    def __ne__(self, right):
        return not self.__eq__(right)

    @property
    def mono(self):
        u"""A structure that describes the monochromator.

        Attributes
        ----------
        tau : str or float
            The monochromator reciprocal lattice vector in Å\ :sup:`-1`.
            Instead of a numerical input one can use one of the following
            keyword strings:

                +------------------+--------------+-----------+
                | String           |     τ        |           |
                +==================+==============+===========+
                | Be(002)          | 3.50702      |           |
                +------------------+--------------+-----------+
                | Co0.92Fe0.08(200)| 3.54782      | (Heusler) |
                +------------------+--------------+-----------+
                | Cu(002)          | 3.47714      |           |
                +------------------+--------------+-----------+
                | Cu(111)          | 2.99913      |           |
                +------------------+--------------+-----------+
                | Cu(220)          | 4.91642      |           |
                +------------------+--------------+-----------+
                | Cu2MnAl(111)     | 1.82810      | (Heusler) |
                +------------------+--------------+-----------+
                | Ge(111)          | 1.92366      |           |
                +------------------+--------------+-----------+
                | Ge(220)          | 3.14131      |           |
                +------------------+--------------+-----------+
                | Ge(311)          | 3.68351      |           |
                +------------------+--------------+-----------+
                | Ge(511)          | 5.76968      |           |
                +------------------+--------------+-----------+
                | Ge(533)          | 7.28063      |           |
                +------------------+--------------+-----------+
                | PG(002)          | 1.87325      |           |
                +------------------+--------------+-----------+
                | PG(004)          | 3.74650      |           |
                +------------------+--------------+-----------+
                | PG(110)          | 5.49806      |           |
                +------------------+--------------+-----------+
                | Si(111)          | 2.00421      |           |
                +------------------+--------------+-----------+

        mosaic : int
            The monochromator mosaic in minutes of arc.

        vmosaic : int
            The vertical mosaic of monochromator in minutes of arc. If
            this field is left unassigned, an isotropic mosaic is assumed.

        dir : int
            Direction of the crystal (left or right, -1 or +1, respectively).
            Default: -1 (left-handed coordinate frame).

        rh : float
            Horizontal curvature of the monochromator in cm.

        rv : float
            Vertical curvature of the monochromator in cm.

        """
        return self._mono

    @mono.setter
    def mono(self, value):
        self._mono = value

    @property
    def ana(self):
        u"""A structure that describes the analyzer and contains fields as in
        :attr:`mono` plus optional fields.

        Attributes
        ----------
        thickness: float
            The analyzer thickness in cm for ideal-crystal reflectivity
            corrections (Section II C 3). If no reflectivity corrections are to
            be made, this field should remain unassigned or set to a negative
            value.

        Q : float
            The kinematic reflectivity coefficient for this correction. It is
            given by

            .. math::    Q = \\frac{4|F|**2}{V_0} \\frac{(2\\pi)**3}{\\tau**3},

            where V0 is the unit cell volume for the analyzer crystal, F is the
            structure factor of the analyzer reflection, and τ is the analyzer
            reciprocal lattice vector. For PG(002) Q = 0.1287. Leave this field
            unassigned or make it negative if you don’t want the correction
            done.

        horifoc : bool
            A flag that is set to 1 if a horizontally focusing analyzer is used
            (Section II D). In this case ``hcol[2]`` (see below) is the angular
            size of the analyzer, as seen from the sample position. If the
            field is unassigned or equal to -1, a flat analyzer is assumed.
            Note that this option is only available with the Cooper-Nathans
            method.

        dir : int
            Direction of the crystal (left or right, -1 or +1, respectively).
            Default: -1 (left-handed coordinate frame).

        rh : float
            Horizontal curvature of the analyzer in cm.

        rv : float
            Vertical curvature of the analyzer in cm.

        """
        return self._ana

    @ana.setter
    def ana(self, value):
        self._ana = value

    @property
    def method(self):
        """Selects the computation method.
        If ``method=0`` or left undefined, a Cooper-Nathans calculation is
        performed. For a Popovici calculation set ``method=1``.
        """
        return self._method

    @method.setter
    def method(self, value):
        self._method = value

    @property
    def moncor(self):
        """Selects the type of normalization used to calculate ``R0``
        If ``moncor=1`` or left undefined, ``R0`` is calculated in
        normalization to monitor counts (Section II C 2). 1/k\ :sub:`i` monitor
        efficiency correction is included automatically. To normalize ``R0`` to
        source flux (Section II C 1), use ``moncor=0``.
        """
        return self._moncar

    @moncor.setter
    def moncor(self, value):
        self._moncar = value

    @property
    def hcol(self):
        r""" The horizontal Soller collimations in minutes of arc (FWHM beam
        divergence) starting from the in-pile collimator. In case of a
        horizontally-focusing analyzer ``hcol[2]`` is the angular size of the
        analyzer, as seen from the sample position. If the beam divergence is
        limited by a neutron guide, the corresponding element of :attr:`hcol`
        is the negative of the guide’s *m*-value. For example, for a 58-Ni
        guide ( *m* = 1.2 ) before the monochromator, ``hcol[0]`` should be
        -1.2.
        """
        return self._hcol

    @hcol.setter
    def hcol(self, value):
        self._hcol = value

    @property
    def vcol(self):
        """The vertical Soller collimations in minutes of arc (FWHM beam
        divergence) starting from the in-pile collimator. If the beam
        divergence is limited by a neutron guide, the corresponding element of
        :attr:`vcol` is the negative of the guide’s *m*-value. For example, for
        a 58-Ni guide ( *m* = 1.2 ) before the monochromator, ``vcol[0]``
        should be -1.2.
        """
        return self._vcol

    @vcol.setter
    def vcol(self, value):
        self._vcol = value

    @property
    def arms(self):
        """distances between the source and monochromator, monochromator
        and sample, sample and analyzer, analyzer and detector, and
        monochromator and monitor, respectively. The 5th element is only needed
        if ``moncor=1``
        """
        return self._arms

    @arms.setter
    def arms(self, value):
        self._arms = value

    @property
    def efixed(self):
        """the fixed incident or final neutron energy, in meV.
        """
        return self._efixed

    @efixed.setter
    def efixed(self, value):
        self._efixed = value

    @property
    def sample(self):
        """A structure that describes the sample.

        Attributes
        ----------
        mosaic
            FWHM sample mosaic in the scattering plane
            in minutes of arc. If left unassigned, no sample
            mosaic corrections (section II E) are performed.

        vmosaic
            The vertical sample mosaic in minutes of arc.
            If left unassigned, isotropic mosaic is assumed.

        dir
            The direction of the crystal (left or right, -1 or +1,
            respectively). Default: -1 (left-handed coordinate frame).

        """
        return self._sample

    @sample.setter
    def sample(self, value):
        self._sample = value

    @property
    def orient1(self):
        """Miller indexes of the first reciprocal-space orienting vector for
        the S coordinate system, as explained in Section II G.
        """
        return self._sample.u

    @orient1.setter
    def orient1(self, value):
        self._sample.u = np.array(value)

    @property
    def orient2(self):
        """Miller indexes of the second reciprocal-space orienting vector
        for the S coordinate system, as explained in Section II G.
        """
        return self._sample.v

    @orient2.setter
    def orient2(self, value):
        self._sample.v = np.array(value)

    @property
    def infin(self):
        """a flag set to -1 or left unassigned if the final energy is fixed, or
        set to +1 in a fixed-incident setup.
        """
        return self._infin

    @infin.setter
    def infin(self, value):
        self._infin = value

    @property
    def guide(self):
        r"""A structure that describes the source
        """
        return self._guide

    @guide.setter
    def guide(self, value):
        self._guide = value

    @property
    def detector(self):
        """A structure that describes the detector
        """
        return self._detector

    @detector.setter
    def detector(self, value):
        self._detector = value

    @property
    def monitor(self):
        """A structure that describes the monitor
        """
        return self._monitor

    @monitor.setter
    def monitor(self, value):
        self._monitor = value

    @property
    def Smooth(self):
        u"""Defines the smoothing parameters as explained in Section II H. Leave this
        field unassigned if you don’t want this correction done.

        * ``Smooth.E`` is the smoothing FWHM in energy (meV). A small number
          means “no smoothing along this direction”.

        * ``Smooth.X`` is the smoothing FWHM along the first orienting vector
          (x0 axis) in Å\ :sup:`-1`.

        * ``Smooth.Y`` is the smoothing FWHM along the y axis in Å\ :sup:`-1`.

        * ``Smooth.Z`` is the smoothing FWHM along the vertical direction in
          Å\ :sup:`-1`.

        """
        return self._Smooth

    @Smooth.setter
    def Smooth(self, value):
        self._Smooth = value

    def get_lattice(self):
        r"""Extracts lattice parameters from EXP and returns the direct and
        reciprocal lattice parameters in the form used by _scalar.m, _star.m,
        etc.

        Returns
        -------
        [lattice, rlattice] : [class, class]
            Returns the direct and reciprocal lattice sample classes

        Notes
        -----
        Translated from ResLib 3.4c, originally authored by A. Zheludev,
        1999-2007, Oak Ridge National Laboratory

        """
        lattice = Sample(self.sample.a,
                         self.sample.b,
                         self.sample.c,
                         np.deg2rad(self.sample.alpha),
                         np.deg2rad(self.sample.beta),
                         np.deg2rad(self.sample.gamma))
        rlattice = _star(lattice)[-1]

        return [lattice, rlattice]

    def _StandardSystem(self):
        r"""Returns rotation matrices to calculate resolution in the sample view
        instead of the instrument view

        Attributes
        ----------
        EXP : class
            Instrument class

        Returns
        -------
        [x, y, z, lattice, rlattice] : [array, array, array, class, class]
            Returns the rotation matrices and real and reciprocal lattice
            sample classes

        Notes
        -----
        Translated from ResLib 3.4c, originally authored by A. Zheludev,
        1999-2007, Oak Ridge National Laboratory

        """
        [lattice, rlattice] = self.get_lattice()

        orient1 = self.orient1
        orient2 = self.orient2

        modx = _modvec(orient1, rlattice)

        x = orient1 / modx

        proj = _scalar(orient2, x, rlattice)

        y = orient2 - x * proj

        mody = _modvec(y, rlattice)

        if len(np.where(mody <= 0)[0]) > 0:
            raise ScatteringTriangleError('Orienting vectors are colinear')

        y /= mody

        z = np.array([ x[1] * y[2] - y[1] * x[2],
                       x[2] * y[0] - y[2] * x[0],
                      -x[1] * y[0] + y[1] * x[0]], dtype=np.float64)

        proj = _scalar(z, x, rlattice)

        z -= x * proj

        proj = _scalar(z, y, rlattice)

        z -= y * proj

        modz = _modvec(z, rlattice)

        z /= modz

        return [x, y, z, lattice, rlattice]

    def calc_resolution_in_Q_coords(self, Q, W):
        r"""For a momentum transfer Q and energy transfers W, given experimental
        conditions specified in EXP, calculates the Cooper-Nathans or Popovici
        resolution matrix RM and resolution prefactor R0 in the Q coordinate
        system (defined by the scattering vector and the scattering plane).

        Parameters
        ----------
        Q : ndarray or list of ndarray
            The Q vectors in reciprocal space at which resolution should be
            calculated, in inverse angstroms

        W : float or list of floats
            The energy transfers at which resolution should be calculated in meV

        Returns
        -------
        [R0, RM] : list(float, ndarray)
            Resolution pre-factor (R0) and resolution matrix (RM) at the given
            reciprocal lattice vectors and energy transfers

        Notes
        -----
        Translated from ResLib 3.4c, originally authored by A. Zheludev,
        1999-2007, Oak Ridge National Laboratory

        """
        CONVERT1 = np.pi / 60. / 180. / np.sqrt(8 * np.log(2))
        CONVERT2 = 2.072

        [length, Q, W] = _CleanArgs(Q, W)

        RM = np.zeros((length, 4, 4), dtype=np.float64)
        R0 = np.zeros(length, dtype=np.float64)
        RM_ = np.zeros((4, 4), dtype=np.float64)

        # the method to use
        method = 0
        if hasattr(self, 'method'):
            method = self.method

        # Assign default values and decode parameters
        moncor = 1
        if hasattr(self, 'moncor'):
            moncor = self.moncor

        alpha = np.array(self.hcol) * CONVERT1
        beta = np.array(self.vcol) * CONVERT1
        mono = self.mono
        etam = np.array(mono.mosaic) * CONVERT1
        etamv = np.copy(etam)
        if hasattr(mono, 'vmosaic') and (method == 1 or method == 'Popovici'):
            etamv = np.array(mono.vmosaic) * CONVERT1

        ana = self.ana
        etaa = np.array(ana.mosaic) * CONVERT1
        etaav = np.copy(etaa)
        if hasattr(ana, 'vmosaic'):
            etaav = np.array(ana.vmosaic) * CONVERT1

        sample = self.sample
        infin = -1
        if hasattr(self, 'infin'):
            infin = self.infin

        efixed = self.efixed

        monitorw = 1.
        monitorh = 1.
        beamw = 1.
        beamh = 1.
        monow = 1.
        monoh = 1.
        monod = 1.
        anaw = 1.
        anah = 1.
        anad = 1.
        detectorw = 1.
        detectorh = 1.
        sshapes = np.repeat(np.eye(3, dtype=np.float64)[np.newaxis].reshape((1, 3, 3)), length, axis=0)
        sshape_factor = 12.
        L0 = 1.
        L1 = 1.
        L1mon = 1.
        L2 = 1.
        L3 = 1.
        monorv = 1.e6
        monorh = 1.e6
        anarv = 1.e6
        anarh = 1.e6

        if hasattr(self, 'guide'):
            beam = self.guide
            if hasattr(beam, 'width'):
                beamw = beam.width ** 2 / 12.
            if hasattr(beam, 'height'):
                beamh = beam.height ** 2 / 12.
        bshape = np.diag([beamw, beamh])

        if hasattr(self, 'monitor'):
            monitor = self.monitor
            if hasattr(monitor, 'width'):
                monitorw = monitor.width ** 2 / 12.
            monitorh = monitorw
            if hasattr(monitor, 'height'):
                monitorh = monitor.height ** 2 / 12.
        monitorshape = np.diag([monitorw, monitorh])

        if hasattr(self, 'detector'):
            detector = self.detector
            if hasattr(detector, 'width'):
                detectorw = detector.width ** 2 / 12.
            if hasattr(detector, 'height'):
                detectorh = detector.height ** 2 / 12.
        dshape = np.diag([detectorw, detectorh])

        if hasattr(mono, 'width'):
            monow = mono.width ** 2 / 12.
        if hasattr(mono, 'height'):
            monoh = mono.height ** 2 / 12.
        if hasattr(mono, 'depth'):
            monod = mono.depth ** 2 / 12.
        mshape = np.diag([monod, monow, monoh])

        if hasattr(ana, 'width'):
            anaw = ana.width ** 2 / 12.
        if hasattr(ana, 'height'):
            anah = ana.height ** 2 / 12.
        if hasattr(ana, 'depth'):
            anad = ana.depth ** 2 / 12.
        ashape = np.diag([anad, anaw, anah])

        if hasattr(sample, 'shape_type'):
            if sample.shape_type == 'cylindrical':
                sshape_factor = 16.
            elif sample.shape_type == 'rectangular':
                sshape_factor = 12.
        if hasattr(sample, 'width') and hasattr(sample, 'depth') and hasattr(sample, 'height'):
            _sshape = np.diag([sample.depth, sample.width, sample.height]).astype(np.float64) ** 2 / sshape_factor
            sshapes = np.repeat(_sshape[np.newaxis].reshape((1, 3, 3)), length, axis=0)
        elif hasattr(sample, 'shape'):
            _sshape = sample.shape.astype(np.float64) / sshape_factor
            if len(_sshape.shape) == 2:
                sshapes = np.repeat(_sshape[np.newaxis].reshape((1, 3, 3)), length, axis=0)
            else:
                sshapes = _sshape

        if hasattr(self, 'arms') and method == 1:
            arms = self.arms
            L0, L1, L2, L3 = arms[:4]
            L1mon = np.copy(L1)
            if len(arms) > 4:
                L1mon = np.copy(arms[4])

        if hasattr(mono, 'rv'):
            monorv = mono.rv

        if hasattr(mono, 'rh'):
            monorh = mono.rh

        if hasattr(ana, 'rv'):
            anarv = ana.rv

        if hasattr(ana, 'rh'):
            anarh = ana.rh

        taum = GetTau(mono.tau)
        taua = GetTau(ana.tau)

        horifoc = -1
        if hasattr(self, 'horifoc'):
            horifoc = self.horifoc

        if horifoc == 1:
            alpha[2] = alpha[2] * np.sqrt(8. * np.log(2.) / 12.)

        sm = self.mono.dir
        ss = self.sample.dir
        sa = self.ana.dir

        for ind in range(length):
            sshape = sshapes[ind, :, :]
            # Calculate angles and energies
            w = W[ind]
            q = Q[ind]
            ei = efixed
            ef = efixed
            if infin > 0:
                ef = efixed - w
            else:
                ei = efixed + w
            ki = np.sqrt(ei / CONVERT2)
            kf = np.sqrt(ef / CONVERT2)

            thetam = np.arcsin(taum / (2. * ki)) * sm
            thetaa = np.arcsin(taua / (2. * kf)) * sa
            s2theta = np.arccos(np.complex((ki ** 2 + kf ** 2 - q ** 2) / (2. * ki * kf))) * ss
            if np.abs(np.imag(s2theta)) > 1e-12:
                raise ScatteringTriangleError(
                    'KI,KF,Q triangle will not close. Change the value of KFIX,FX,QH,QK or QL.')
            else:
                s2theta = np.real(s2theta)

            # correct sign of curvatures
            monorh = monorh * sm
            monorv = monorv * sm
            anarh = anarh * sa
            anarv = anarv * sa

            thetas = s2theta / 2.
            phi = np.arctan2(-kf * np.sin(s2theta), ki - kf * np.cos(s2theta))

            # Calculate beam divergences defined by neutron guides
            alpha[alpha < 0] = -alpha[alpha < 0] * 0.1 * 60. * (2. * np.pi / ki) / 0.427 / np.sqrt(3.)
            beta[beta < 0] = -beta[beta < 0] * 0.1 * 60. * (2. * np.pi / ki) / 0.427 / np.sqrt(3.)

            # Redefine sample geometry
            psi = thetas - phi  # Angle from sample geometry X axis to Q
            rot = np.matrix([[np.cos(psi), np.sin(psi), 0],
                             [-np.sin(psi), np.cos(psi), 0],
                             [0, 0, 1]], dtype=np.float64)

            # sshape=rot'*sshape*rot
            sshape = np.matrix(rot) * np.matrix(sshape) * np.matrix(rot).H

            # Definition of matrix G
            G = np.matrix(
                np.diag(1. / np.array([alpha[:2], beta[:2], alpha[2:], beta[2:]], dtype=np.float64).flatten() ** 2))

            # Definition of matrix F
            F = np.matrix(np.diag(1. / np.array([etam, etamv, etaa, etaav], dtype=np.float64) ** 2))

            # Definition of matrix A
            A = np.matrix([[ki / 2. / np.tan(thetam), -ki / 2. / np.tan(thetam), 0, 0, 0, 0, 0, 0],
                           [0, ki, 0, 0, 0, 0, 0, 0],
                           [0, 0, 0, ki, 0, 0, 0, 0],
                           [0, 0, 0, 0, kf / 2. / np.tan(thetaa), -kf / 2. / np.tan(thetaa), 0, 0],
                           [0, 0, 0, 0, kf, 0, 0, 0],
                           [0, 0, 0, 0, 0, 0, kf, 0]], dtype=np.float64)

            # Definition of matrix C
            C = np.matrix([[0.5, 0.5, 0, 0, 0, 0, 0, 0],
                           [0., 0., 1. / (2. * np.sin(thetam)), -1. / (2. * np.sin(thetam)), 0, 0, 0, 0],
                           [0, 0, 0, 0, 0.5, 0.5, 0, 0],
                           [0, 0, 0, 0, 0, 0, 1. / (2. * np.sin(thetaa)), -1. / (2. * np.sin(thetaa))]],
                          dtype=np.float64)

            # Definition of matrix Bmatrix
            Bmatrix = np.matrix([[np.cos(phi), np.sin(phi), 0, -np.cos(phi - s2theta), -np.sin(phi - s2theta), 0],
                                 [-np.sin(phi), np.cos(phi), 0, np.sin(phi - s2theta), -np.cos(phi - s2theta), 0],
                                 [0, 0, 1, 0, 0, -1],
                                 [2. * CONVERT2 * ki, 0, 0, -2. * CONVERT2 * kf, 0, 0]], dtype=np.float64)

            # Definition of matrix S
            Sinv = np.matrix(blkdiag(np.array(bshape, dtype=np.float64), mshape, sshape, ashape, dshape))  # S-1 matrix
            S = Sinv.I

            # Definition of matrix T
            T = np.matrix([[-1. / (2. * L0), 0, np.cos(thetam) * (1. / L1 - 1. / L0) / 2.,
                            np.sin(thetam) * (1. / L0 + 1. / L1 - 2. / (monorh * np.sin(thetam))) / 2., 0,
                            np.sin(thetas) / (2. * L1), np.cos(thetas) / (2. * L1), 0, 0, 0, 0, 0, 0],
                           [0, -1. / (2. * L0 * np.sin(thetam)), 0, 0,
                            (1. / L0 + 1. / L1 - 2. * np.sin(thetam) / monorv) / (2. * np.sin(thetam)), 0, 0,
                            -1. / (2. * L1 * np.sin(thetam)), 0, 0, 0, 0, 0],
                           [0, 0, 0, 0, 0, np.sin(thetas) / (2. * L2), -np.cos(thetas) / (2. * L2), 0,
                            np.cos(thetaa) * (1. / L3 - 1. / L2) / 2.,
                            np.sin(thetaa) * (1. / L2 + 1. / L3 - 2. / (anarh * np.sin(thetaa))) / 2., 0,
                            1. / (2. * L3), 0],
                           [0, 0, 0, 0, 0, 0, 0, -1. / (2. * L2 * np.sin(thetaa)), 0, 0,
                            (1. / L2 + 1. / L3 - 2. * np.sin(thetaa) / anarv) / (2. * np.sin(thetaa)), 0,
                            -1. / (2. * L3 * np.sin(thetaa))]], dtype=np.float64)

            # Definition of matrix D
            # Lots of index mistakes in paper for matrix D
            D = np.matrix([[-1. / L0, 0, -np.cos(thetam) / L0, np.sin(thetam) / L0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                           [0, 0, np.cos(thetam) / L1, np.sin(thetam) / L1, 0, np.sin(thetas) / L1, np.cos(thetas) / L1,
                            0, 0, 0, 0, 0, 0],
                           [0, -1. / L0, 0, 0, 1. / L0, 0, 0, 0, 0, 0, 0, 0, 0],
                           [0, 0, 0, 0, -1. / L1, 0, 0, 1. / L1, 0, 0, 0, 0, 0],
                           [0, 0, 0, 0, 0, np.sin(thetas) / L2, -np.cos(thetas) / L2, 0, -np.cos(thetaa) / L2,
                            np.sin(thetaa) / L2, 0, 0, 0],
                           [0, 0, 0, 0, 0, 0, 0, 0, np.cos(thetaa) / L3, np.sin(thetaa) / L3, 0, 1. / L3, 0],
                           [0, 0, 0, 0, 0, 0, 0, -1. / L2, 0, 0, 1. / L2, 0, 0],
                           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -1. / L3, 0, 1. / L3]], dtype=np.float64)

            # Definition of resolution matrix M
            if method == 1 or method == 'popovici':
                K = S + T.H * F * T
                H = np.linalg.inv(D * np.linalg.inv(K) * D.H)
                Ninv = A * np.linalg.inv(H + G) * A.H
            else:
                H = G + C.H * F * C
                Ninv = A * np.linalg.inv(H) * A.H
                # Horizontally focusing analyzer if needed
                if horifoc > 0:
                    Ninv = np.linalg.inv(Ninv)
                    Ninv[3:5, 3:5] = np.matrix([[(np.tan(thetaa) / (etaa * kf)) ** 2, 0],
                                                [0, (1 / (kf * alpha[2])) ** 2]], dtype=np.float64)
                    Ninv = np.linalg.inv(Ninv)

            Minv = Bmatrix * Ninv * Bmatrix.H

            M = np.linalg.inv(Minv)
            RM_ = np.copy(M)

            # Calculation of prefactor, normalized to source
            Rm = ki ** 3 / np.tan(thetam)
            Ra = kf ** 3 / np.tan(thetaa)
            R0_ = Rm * Ra * (2. * np.pi) ** 4 / (64. * np.pi ** 2 * np.sin(thetam) * np.sin(thetaa))

            if method == 1 or method == 'popovici':
                # Popovici
                R0_ = R0_ * np.sqrt(np.linalg.det(F) / np.linalg.det(H + G))
            else:
                # Cooper-Nathans (popovici Eq 5 and 9)
                R0_ = R0_ * np.sqrt(np.linalg.det(F) / np.linalg.det(H))

            # Normalization to flux on monitor
            if moncor == 1:
                g = G[:4, :4]
                f = F[:2, :2]
                c = C[:2, :4]

                t = np.matrix([[-1. / (2. * L0), 0, np.cos(thetam) * (1. / L1mon - 1. / L0) / 2.,
                                np.sin(thetam) * (1. / L0 + 1. / L1mon - 2. / (monorh * np.sin(thetam))) / 2., 0, 0,
                                1. / (2. * L1mon)],
                               [0, -1. / (2. * L0 * np.sin(thetam)), 0, 0,
                                (1. / L0 + 1. / L1mon - 2. * np.sin(thetam) / monorv) / (2. * np.sin(thetam)), 0, 0]],
                              dtype=np.float64)

                sinv = blkdiag(np.array(bshape, dtype=np.float64), mshape, monitorshape)  # S-1 matrix
                s = np.linalg.inv(sinv)

                d = np.matrix([[-1. / L0, 0, -np.cos(thetam) / L0, np.sin(thetam) / L0, 0, 0, 0],
                               [0, 0, np.cos(thetam) / L1mon, np.sin(thetam) / L1mon, 0, 0, 1. / L1mon],
                               [0, -1. / L0, 0, 0, 1. / L0, 0, 0],
                               [0, 0, 0, 0, -1. / L1mon, 0, 0]], dtype=np.float64)

                if method == 1 or method == 'popovici':
                    # Popovici
                    Rmon = Rm * (2 * np.pi) ** 2 / (8 * np.pi * np.sin(thetam)) * np.sqrt(
                        np.linalg.det(f) / np.linalg.det(np.linalg.inv(d * np.linalg.inv(s + t.H * f * t) * d.H) + g))
                else:
                    # Cooper-Nathans
                    Rmon = Rm * (2 * np.pi) ** 2 / (8 * np.pi * np.sin(thetam)) * np.sqrt(
                        np.linalg.det(f) / np.linalg.det(g + c.H * f * c))

                R0_ = R0_ / Rmon
                R0_ = R0_ * ki  # 1/ki monitor efficiency

            # Transform prefactor to Chesser-Axe normalization
            R0_ = R0_ / (2. * np.pi) ** 2 * np.sqrt(np.linalg.det(RM_))
            # Include kf/ki part of cross section
            R0_ = R0_ * kf / ki

            # Take care of sample mosaic if needed
            # [S. A. Werner & R. Pynn, J. Appl. Phys. 42, 4736, (1971), eq 19]
            if hasattr(sample, 'mosaic'):
                etas = sample.mosaic * CONVERT1
                etasv = np.copy(etas)
                if hasattr(sample, 'vmosaic'):
                    etasv = sample.vmosaic * CONVERT1
                R0_ = R0_ / np.sqrt((1 + (q * etas) ** 2 * RM_[2, 2]) * (1 + (q * etasv) ** 2 * RM_[1, 1]))
                Minv[1, 1] = Minv[1, 1] + q ** 2 * etas ** 2
                Minv[2, 2] = Minv[2, 2] + q ** 2 * etasv ** 2
                RM_ = np.linalg.inv(Minv)

            # Take care of analyzer reflectivity if needed [I. Zaliznyak, BNL]
            if hasattr(ana, 'thickness') and hasattr(ana, 'Q'):
                KQ = ana.Q
                KT = ana.thickness
                toa = (taua / 2.) / np.sqrt(kf ** 2 - (taua / 2.) ** 2)
                smallest = alpha[3]
                if alpha[3] > alpha[2]:
                    smallest = alpha[2]
                Qdsint = KQ * toa
                dth = (np.arange(1, 201) / 200.) * np.sqrt(2. * np.log(2.)) * smallest
                wdth = np.exp(-dth ** 2 / 2. / etaa ** 2)
                sdth = KT * Qdsint * wdth / etaa / np.sqrt(2. * np.pi)
                rdth = 1. / (1 + 1. / sdth)
                reflec = sum(rdth) / sum(wdth)
                R0_ = R0_ * reflec

            R0[ind] = R0_
            RM[ind] = RM_.copy()

        return [R0, RM]

    def calc_resolution(self, hkle):
        r"""For a scattering vector (H,K,L) and  energy transfers W, given
        experimental conditions specified in EXP, calculates the Cooper-Nathans
        resolution matrix RMS and Cooper-Nathans Resolution prefactor R0 in a
        coordinate system defined by the crystallographic axes of the sample.

        Parameters
        ----------
        hkle : list
            Array of the scattering vector and energy transfer at which the
            calculation should be performed

        Notes
        -----
            Translated from ResLib, originally authored by A. Zheludev, 1999-2007,
            Oak Ridge National Laboratory

        """
        self.HKLE = hkle
        [H, K, L, W] = hkle

        [length, H, K, L, W] = _CleanArgs(H, K, L, W)
        self.H, self.K, self.L, self.W = H, K, L, W

        [x, y, z, sample, rsample] = self._StandardSystem()
        del z, sample

        Q = _modvec([H, K, L], rsample)
        uq = np.vstack((H / Q, K / Q, L / Q))

        xq = _scalar(x, uq, rsample)
        yq = _scalar(y, uq, rsample)

        tmat = np.array(
            [np.array([[xq[i], yq[i], 0, 0], [-yq[i], xq[i], 0, 0], [0, 0, 1., 0], [0, 0, 0, 1.]], dtype=np.float64) for i in range(len(xq))])

        RMS = np.zeros((length, 4, 4), dtype=np.float64)
        rot = np.zeros((3, 3), dtype=np.float64)

        # Sample shape matrix in coordinate system defined by scattering vector
        sample = self.sample
        if hasattr(sample, 'shape'):
            samples = []
            for i in range(length):
                rot = tmat[i, :3, :3]
                samples.append(np.matrix(rot) * np.matrix(sample.shape) * np.matrix(rot).H)
            self.sample.shape = np.array(samples)

        [R0, RM] = self.calc_resolution_in_Q_coords(Q, W)

        for i in range(length):
            RMS[i] = np.matrix(tmat[i]).H * np.matrix(RM[i]) * np.matrix(tmat[i])

        e = np.identity(4)
        for i in range(length):
            if hasattr(self, 'Smooth'):
                if self.Smooth.X:
                    mul = np.diag([1 / (self.Smooth.X ** 2 / 8 / np.log(2)),
                                   1 / (self.Smooth.Y ** 2 / 8 / np.log(2)),
                                   1 / (self.Smooth.E ** 2 / 8 / np.log(2)),
                                   1 / (self.Smooth.Z ** 2 / 8 / np.log(2))])
                    R0[i] = R0[i] / np.sqrt(np.linalg.det(np.matrix(e) / np.matrix(RMS[i]))) * np.sqrt(
                        np.linalg.det(np.matrix(e) / np.matrix(mul) + np.matrix(e) / np.matrix(RMS[i])))
                    RMS[i] = np.matrix(e) / (
                        np.matrix(e) / np.matrix(mul) + np.matrix(e) / np.matrix(RMS[i]))

        self.R0, self.RMS, self.RM = [np.squeeze(item) for item in (R0, RMS, RM)]

    def get_angles_and_Q(self, hkle):
        r"""Returns the Triple Axis Spectrometer angles and Q-vector given
        position in reciprocal space

        Parameters
        ----------
        hkle : list
            Array of the scattering vector and energy transfer at which the
            calculation should be performed

        Returns
        -------
        [A, Q] : list
            The angles A (A1 -- A5 in a list of floats) and Q (ndarray)

        """
        # compute all TAS angles (in plane)

        h, k, l, w = hkle
        # compute angles
        try:
            fx = 2 * int(self.infin == -1) + int(self.infin == 1)
        except AttributeError:
            fx = 2

        kfix = Energy(energy=self.efixed).wavevector
        f = 0.4826  # f converts from energy units into k^2, f=0.4826 for meV
        ki = np.sqrt(kfix ** 2 + (fx - 1) * f * w)  # kinematical equations.
        kf = np.sqrt(kfix ** 2 - (2 - fx) * f * w)

        # compute the transversal Q component, and A3 (sample rotation)
        # from McStas templateTAS.instr and TAS MAD ILL
        a = np.array([self.sample.a, self.sample.b, self.sample.c]) / (2 * np.pi)
        alpha = np.deg2rad([self.sample.alpha, self.sample.beta, self.sample.gamma])
        cosa = np.cos(alpha)
        sina = np.sin(alpha)
        cc = np.sum(cosa * cosa)
        cc = 1 + 2 * np.product(cosa) - cc
        cc = np.sqrt(cc)
        b = sina / (a * cc)
        c1 = np.roll(cosa[np.newaxis].T, -1)
        c2 = np.roll(c1, -1)
        s1 = np.roll(sina[np.newaxis].T, -1)
        s2 = np.roll(s1, -1)
        cosb = (c1 * c2 - cosa[np.newaxis].T) / (s1 * s2)
        sinb = np.sqrt(1 - cosb * cosb)

        bb = np.array([[b[0], 0, 0],
                       [b[1] * cosb[2], b[1] * sinb[2], 0],
                       [b[2] * cosb[1], -b[2] * sinb[1] * cosa[0], 1 / a[2]]])
        bb = bb.T

        aspv = np.hstack((self.orient1[np.newaxis].T, self.orient2[np.newaxis].T))

        vv = np.zeros((3, 3))
        vv[0:2, :] = np.transpose(np.dot(bb, aspv))
        for m in range(2, 0, -1):
            vt = np.roll(np.roll(vv, -1, axis=0), -1, axis=1) * np.roll(np.roll(vv, -2, axis=0), -2, axis=1) - np.roll(
                np.roll(vv, -1, axis=0), -2, axis=1) * np.roll(np.roll(vv, -2, axis=0), -1, axis=1)
            vv[m, :] = vt[m, :]

        c = np.sqrt(np.sum(vv * vv, axis=0))

        vv = vv / np.tile(c, (3, 1))
        s = vv.T * bb

        qt = np.squeeze(np.dot(np.array([h, k, l]).T, s.T))

        qs = np.sum(qt ** 2)
        Q = np.sqrt(qs)

        sm = self.mono.dir
        ss = self.sample.dir
        sa = self.ana.dir
        dm = 2 * np.pi / GetTau(self.mono.tau)
        da = 2 * np.pi / GetTau(self.ana.tau)
        thetaa = sa * np.arcsin(np.pi / (da * kf))  # theta angles for analyser
        thetam = sm * np.arcsin(np.pi / (dm * ki))  # and monochromator.
        thetas = ss * 0.5 * np.arccos((ki ** 2 + kf ** 2 - Q ** 2) / (2 * ki * kf))  # scattering angle from sample.

        A3 = -np.arctan2(qt[1], qt[0]) - np.arccos(
            (np.dot(kf, kf) - np.dot(Q, Q) - np.dot(ki, ki)) / (-2 * np.dot(Q, ki)))
        A3 = ss * A3

        A1 = thetam
        A2 = 2 * A1
        A4 = 2 * thetas
        A5 = thetaa
        A6 = 2 * A5

        A = np.squeeze(np.rad2deg([A1, A2, A3, A4, A5, A6]))

        return [A, Q]


    def resolution_convolution(self, sqw, pref, nargout, hkle, METHOD='fix', ACCURACY=None, p=None, seed=None):
        r"""Numerically calculate the convolution of a user-defined
        cross-section function with the resolution function for a
        3-axis neutron scattering experiment.

        Parameters
        ----------
        sqw : func
            User-supplied "fast" model cross section.

        pref : func
            User-supplied "slow" cross section prefactor and background
            function.

        nargout : int
            Number of arguments returned by the pref function

        hkle : tup
            Tuple of H, K, L, and W, specifying the wave vector and energy
            transfers at which the convolution is to be calculated (i.e.
            define $\mathbf{Q}_0$). H, K, and L are given in reciprocal
            lattice units and W in meV.

        EXP : obj
            Instrument object containing all information on experimental setup.

        METHOD : str
            Specifies which 4D-integration method to use. 'fix' (Default):
            sample the cross section on a fixed grid of points uniformly
            distributed $\phi$-space. 2*ACCURACY[0]+1 points are sampled
            along $\phi_1$, $\phi_2$, and $\phi_3$, and 2*ACCURACY[1]+1
            along $\phi_4$ (vertical direction). 'mc': 4D Monte Carlo
            integration. The cross section is sampled in 1000*ACCURACY
            randomly chosen points, uniformly distributed in $\phi$-space.

        ACCURACY : array(2) or int
            Determines the number of sampling points in the integration.

        p : list
            A parameter that is passed on, without change to sqw and pref.

        Returns
        -------
        conv : array
            Calculated value of the cross section, folded with the resolution
            function at the given $\mathbf{Q}_0$

        Notes
        -----
        Translated from ResLib 3.4c, originally authored by A. Zheludev,
        1999-2007, Oak Ridge National Laboratory

        """
        self.calc_resolution(hkle)
        [R0, RMS] = [np.copy(self.R0), self.RMS.copy()]

        H, K, L, W = hkle
        [length, H, K, L, W] = _CleanArgs(H, K, L, W)
        [xvec, yvec, zvec] = self._StandardSystem()[:3]

        Mxx = RMS[:, 0, 0]
        Mxy = RMS[:, 0, 1]
        Mxw = RMS[:, 0, 3]
        Myy = RMS[:, 1, 1]
        Myw = RMS[:, 1, 3]
        Mzz = RMS[:, 2, 2]
        Mww = RMS[:, 3, 3]

        Mxx -= Mxw ** 2. / Mww
        Mxy -= Mxw * Myw / Mww
        Myy -= Myw ** 2. / Mww
        MMxx = Mxx - Mxy ** 2. / Myy

        detM = MMxx * Myy * Mzz * Mww

        tqz = 1. / np.sqrt(Mzz)
        tqx = 1. / np.sqrt(MMxx)
        tqyy = 1. / np.sqrt(Myy)
        tqyx = -Mxy / Myy / np.sqrt(MMxx)
        tqww = 1. / np.sqrt(Mww)
        tqwy = -Myw / Mww / np.sqrt(Myy)
        tqwx = -(Mxw / Mww - Myw / Mww * Mxy / Myy) / np.sqrt(MMxx)

        inte = sqw(H, K, L, W, p)
        [modes, points] = inte.shape

        if pref is None:
            prefactor = np.ones((modes, points))
            bgr = 0
        else:
            if nargout == 2:
                [prefactor, bgr] = pref(H, K, L, W, self, p)
            elif nargout == 1:
                prefactor = pref(H, K, L, W, self, p)
                bgr = 0
            else:
                raise ValueError('Invalid number or output arguments in prefactor function')

        if METHOD == 'fix':
            if ACCURACY is None:
                ACCURACY = np.array([7, 0])
            M = ACCURACY
            step1 = np.pi / (2 * M[0] + 1)
            step2 = np.pi / (2 * M[1] + 1)
            dd1 = np.linspace(-np.pi / 2 + step1 / 2, np.pi / 2 - step1 / 2, (2 * M[0] + 1))
            dd2 = np.linspace(-np.pi / 2 + step2 / 2, np.pi / 2 - step2 / 2, (2 * M[1] + 1))
            convs = np.zeros((modes, length))
            conv = np.zeros(length)
            [cw, cx, cy] = np.meshgrid(dd1, dd1, dd1, indexing='ij')
            tx = np.tan(cx)
            ty = np.tan(cy)
            tw = np.tan(cw)
            tz = np.tan(dd2)
            norm = np.exp(-0.5 * (tx ** 2 + ty ** 2)) * (1 + tx ** 2) * (1 + ty ** 2) * np.exp(-0.5 * (tw ** 2)) * (
                1 + tw ** 2)
            normz = np.exp(-0.5 * (tz ** 2)) * (1 + tz ** 2)

            for iz in range(len(tz)):
                for i in range(length):
                    dQ1 = tqx[i] * tx
                    dQ2 = tqyy[i] * ty + tqyx[i] * tx
                    dW = tqwx[i] * tx + tqwy[i] * ty + tqww[i] * tw
                    dQ4 = tqz[i] * tz[iz]
                    H1 = H[i] + dQ1 * xvec[0] + dQ2 * yvec[0] + dQ4 * zvec[0]
                    K1 = K[i] + dQ1 * xvec[1] + dQ2 * yvec[1] + dQ4 * zvec[1]
                    L1 = L[i] + dQ1 * xvec[2] + dQ2 * yvec[2] + dQ4 * zvec[2]
                    W1 = W[i] + dW
                    inte = sqw(H1, K1, L1, W1, p)
                    for j in range(modes):
                        add = inte[j, :] * norm * normz[iz]
                        convs[j, i] = convs[j, i] + np.sum(add)

                    conv[i] = np.sum(convs[:, i] * prefactor[:, i])

            conv = conv * step1 ** 3 * step2 / np.sqrt(detM)
            if M[1] == 0:
                conv *= 0.79788
            if M[0] == 0:
                conv *= 0.79788 ** 3

        elif METHOD == 'mc':
            if isinstance(ACCURACY, (list, np.ndarray, tuple)):
                if len(ACCURACY) == 1:
                    ACCURACY = ACCURACY[0]
                else:
                    raise ValueError('ACCURACY must be an int when using Monte Carlo method: {0}'.format(ACCURACY))
            if ACCURACY is None:
                ACCURACY = 10
            M = ACCURACY
            convs = np.zeros((modes, length))
            conv = np.zeros(length)
            for i in range(length):
                for MonteCarlo in range(M):
                    if seed is not None:
                        np.random.seed(seed)
                    r = np.random.randn(4, 1000) * np.pi - np.pi / 2
                    cx = r[0, :]
                    cy = r[1, :]
                    cz = r[2, :]
                    cw = r[3, :]
                    tx = np.tan(cx)
                    ty = np.tan(cy)
                    tz = np.tan(cz)
                    tw = np.tan(cw)
                    norm = np.exp(-0.5 * (tx ** 2 + ty ** 2 + tz ** 2 + tw ** 2)) * (1 + tx ** 2) * (1 + ty ** 2) * (
                        1 + tz ** 2) * (1 + tw ** 2)
                    dQ1 = tqx[i] * tx
                    dQ2 = tqyy[i] * ty + tqyx[i] * tx
                    dW = tqwx[i] * tx + tqwy[i] * ty + tqww[i] * tw
                    dQ4 = tqz[i] * tz
                    H1 = H[i] + dQ1 * xvec[0] + dQ2 * yvec[0] + dQ4 * zvec[0]
                    K1 = K[i] + dQ1 * xvec[1] + dQ2 * yvec[1] + dQ4 * zvec[1]
                    L1 = L[i] + dQ1 * xvec[2] + dQ2 * yvec[2] + dQ4 * zvec[2]
                    W1 = W[i] + dW
                    inte = sqw(H1, K1, L1, W1, p)
                    for j in range(modes):
                        add = inte[j, :] * norm
                        convs[j, i] = convs[j, i] + np.sum(add)

                    conv[i] = np.sum(convs[:, i] * prefactor[:, i])

            conv = conv / M / 1000 * np.pi ** 4. / np.sqrt(detM)

        else:
            raise ValueError('Unknown METHOD: {0}. Valid options are: "fix",  "mc"'.format(METHOD))

        conv *= R0
        conv += bgr

        return conv

    def resolution_convolution_SMA(self, sqw, pref, nargout, hkle, METHOD='fix', ACCURACY=None, p=None, seed=None):
        r"""Numerically calculate the convolution of a user-defined single-mode
        cross-section function with the resolution function for a 3-axis
        neutron scattering experiment.

        Parameters
        ----------
        sqw : func
            User-supplied "fast" model cross section.

        pref : func
            User-supplied "slow" cross section prefactor and background
            function.

        nargout : int
            Number of arguments returned by the pref function

        hkle : tup
            Tuple of H, K, L, and W, specifying the wave vector and energy
            transfers at which the convolution is to be calculated (i.e.
            define $\mathbf{Q}_0$). H, K, and L are given in reciprocal
            lattice units and W in meV.

        EXP : obj
            Instrument object containing all information on experimental setup.

        METHOD : str
            Specifies which 3D-integration method to use. 'fix' (Default):
            sample the cross section on a fixed grid of points uniformly
            distributed $\phi$-space. 2*ACCURACY[0]+1 points are sampled
            along $\phi_1$, and $\phi_2$, and 2*ACCURACY[1]+1 along $\phi_3$
            (vertical direction). 'mc': 3D Monte Carlo integration. The cross
            section is sampled in 1000*ACCURACY randomly chosen points,
            uniformly distributed in $\phi$-space.

        ACCURACY : array(2) or int
            Determines the number of sampling points in the integration.

        p : list
            A parameter that is passed on, without change to sqw and pref.

        Returns
        -------
        conv : array
            Calculated value of the cross section, folded with the resolution
            function at the given $\mathbf{Q}_0$

        Notes
        -----
        Translated from ResLib 3.4c, originally authored by A. Zheludev,
        1999-2007, Oak Ridge National Laboratory

        """
        self.calc_resolution(hkle)
        [R0, RMS] = [np.copy(self.R0), self.RMS.copy()]

        H, K, L, W = hkle
        [length, H, K, L, W] = _CleanArgs(H, K, L, W)

        [xvec, yvec, zvec] = self._StandardSystem()[:3]

        Mww = RMS[:, 3, 3]
        Mxw = RMS[:, 0, 3]
        Myw = RMS[:, 1, 3]

        GammaFactor = np.sqrt(Mww / 2)
        OmegaFactorx = Mxw / np.sqrt(2 * Mww)
        OmegaFactory = Myw / np.sqrt(2 * Mww)

        Mzz = RMS[:, 2, 2]
        Mxx = RMS[:, 0, 0]
        Mxx -= Mxw ** 2 / Mww
        Myy = RMS[:, 1, 1]
        Myy -= Myw ** 2 / Mww
        Mxy = RMS[:, 0, 1]
        Mxy -= Mxw * Myw / Mww

        detxy = np.sqrt(Mxx * Myy - Mxy ** 2)
        detz = np.sqrt(Mzz)

        tqz = 1. / detz
        tqy = np.sqrt(Mxx) / detxy
        tqxx = 1. / np.sqrt(Mxx)
        tqxy = Mxy / np.sqrt(Mxx) / detxy

        [disp, inte] = sqw(H, K, L, p)[:2]
        [modes, points] = disp.shape

        if pref is None:
            prefactor = np.ones(modes, points)
            bgr = 0
        else:
            if nargout == 2:
                [prefactor, bgr] = pref(H, K, L, W, self, p)
            elif nargout == 1:
                prefactor = pref(H, K, L, W, self, p)
                bgr = 0
            else:
                raise ValueError('Invalid number or output arguments in prefactor function')

        if METHOD == 'mc':
            if isinstance(ACCURACY, (list, np.ndarray, tuple)):
                if len(ACCURACY) == 1:
                    ACCURACY = ACCURACY[0]
                else:
                    raise ValueError("ACCURACY (type: {0}) must be an 'int' when using Monte Carlo method".format(type(ACCURACY)))
            if ACCURACY is None:
                ACCURACY = 10
            M = ACCURACY
            convs = np.zeros((modes, length))
            conv = np.zeros(length)
            for i in range(length):
                for MonteCarlo in range(M):
                    if seed is not None:
                        np.random.seed(seed)
                    r = np.random.randn(3, 1000) * np.pi - np.pi / 2
                    cx = r[0, :]
                    cy = r[1, :]
                    cz = r[2, :]
                    tx = np.tan(cx)
                    ty = np.tan(cy)
                    tz = np.tan(cz)
                    norm = np.exp(-0.5 * (tx ** 2 + ty ** 2 + tz ** 2)) * (1 + tx ** 2) * (1 + ty ** 2) * (1 + tz ** 2)
                    dQ1 = tqxx[i] * tx - tqxy[i] * ty
                    dQ2 = tqy[i] * ty
                    dQ4 = tqz[i] * tz
                    H1 = H[i] + dQ1 * xvec[0] + dQ2 * yvec[0] + dQ4 * zvec[0]
                    K1 = K[i] + dQ1 * xvec[1] + dQ2 * yvec[1] + dQ4 * zvec[1]
                    L1 = L[i] + dQ1 * xvec[2] + dQ2 * yvec[2] + dQ4 * zvec[2]
                    [disp, inte, WL] = sqw(H1, K1, L1, p)
                    [modes, points] = disp.shape
                    for j in range(modes):
                        Gamma = WL[j, :] * GammaFactor[i]
                        Omega = GammaFactor[i] * (disp[j, :] - W[i]) + OmegaFactorx[i] * dQ1 + OmegaFactory[i] * dQ2
                        add = inte[j, :] * _voigt(Omega, Gamma) * norm / detxy[i] / detz[i]
                        convs[j, i] = convs[j, i] + np.sum(add)

                conv[i] = np.sum(convs[:, i] * prefactor[:, i])

            conv = conv / M / 1000. * np.pi ** 3

        elif METHOD == 'fix':
            if ACCURACY is None:
                ACCURACY = [7, 0]
            M = ACCURACY
            step1 = np.pi / (2 * M[0] + 1)
            step2 = np.pi / (2 * M[1] + 1)
            dd1 = np.linspace(-np.pi / 2 + step1 / 2, np.pi / 2 - step1 / 2, (2 * M[0] + 1))
            dd2 = np.linspace(-np.pi / 2 + step2 / 2, np.pi / 2 - step2 / 2, (2 * M[1] + 1))
            convs = np.zeros((modes, length))
            conv = np.zeros(length)
            [cy, cx] = np.meshgrid(dd1, dd1, indexing='ij')
            tx = np.tan(cx.flatten())
            ty = np.tan(cy.flatten())
            tz = np.tan(dd2)
            norm = np.exp(-0.5 * (tx ** 2 + ty ** 2)) * (1 + tx ** 2) * (1 + ty ** 2)
            normz = np.exp(-0.5 * (tz ** 2)) * (1 + tz ** 2)
            for iz in range(tz.size):
                for i in range(length):
                    dQ1 = tqxx[i] * tx - tqxy[i] * ty
                    dQ2 = tqy[i] * ty
                    dQ4 = tqz[i] * tz[iz]
                    H1 = H[i] + dQ1 * xvec[0] + dQ2 * yvec[0] + dQ4 * zvec[0]
                    K1 = K[i] + dQ1 * xvec[1] + dQ2 * yvec[1] + dQ4 * zvec[1]
                    L1 = L[i] + dQ1 * xvec[2] + dQ2 * yvec[2] + dQ4 * zvec[2]
                    [disp, inte, WL] = sqw(H1, K1, L1, p)
                    [modes, points] = disp.shape
                    for j in range(modes):
                        Gamma = WL[j, :] * GammaFactor[i]
                        Omega = GammaFactor[i] * (disp[j, :] - W[i]) + OmegaFactorx[i] * dQ1 + OmegaFactory[i] * dQ2
                        add = inte[j, :] * _voigt(Omega, Gamma) * norm * normz[iz] / detxy[i] / detz[i]
                        convs[j, i] = convs[j, i] + np.sum(add)

                    conv[i] = np.sum(convs[:, i] * prefactor[:, i])

            conv = conv * step1 ** 2 * step2

            if M[1] == 0:
                conv *= 0.79788
            if M[0] == 0:
                conv *= 0.79788 ** 2
        else:
            raise ValueError('Unknown METHOD: {0}. Valid options are: "fix" or "mc".'.format(METHOD))

        conv = conv * R0
        conv = conv + bgr

        return conv



from collections import OrderedDict

import numpy as np

from ...data import Data


class DcsMslice(Data):
    r"""Loads DCS_MSLICE (NCNR) exported ascii data files.

    """

    def __init__(self):
        super(DcsMslice, self).__init__()

    def load(self, filename, **kwargs):
        r"""Loads the DCS_MSLICE (NCNR) exported ascii data files, including
        SPE, IEXY, XYE, and XYIE. NOTE: Will NOT load DAVE format files.
        DAVE files are propietary IDL formatted save files.

        Parameters
        ----------
        filename : str
            Path to file to open

        """
        load_file = {'iexy': self.load_iexy,
                     'xyie': self.load_xyie,
                     'spe': self.load_spe,
                     'xye': self.load_xye}

        load_file[self.determine_subtype(filename)](filename)

    def determine_subtype(self, filename):
        try:
            with open(filename) as f:
                first_line = f.readline()
                second_line = f.readline()
                if filename[-4:] == 'iexy' or len(first_line.split()) == 4:
                    return 'iexy'
                elif filename[-4:] == 'xyie' or (len(first_line.split()) == 2 and len(second_line) == 0):
                    return 'xyie'
                elif filename[-3:] == 'spe' or (len(first_line.split()) == 2 and '###' in second_line):
                    return 'spe'
                elif filename[-3:] == 'xye' or len(first_line.split()) == 3:
                    return 'xye'
                else:
                    raise IOError('Cannot determine filetype!')
        except IOError:
            raise

    def load_iexy(self, filename):
        i, e, x, y = np.loadtxt(filename, unpack=True)
        self._data = OrderedDict(intensity=i, error=e, x=x, y=y, monitor=np.ones(len(i)), time=np.ones(len(i)))
        self.data_keys = {'detector': 'intensity', 'monitor': 'monitor', 'time': 'time'}
        self._err = e

    def load_spe(self, filename):
        with open(filename) as f:
            data = []
            for line in f:
                if '###' not in line:
                    _line = line.replace('\n', '').replace('-', ' -').replace('e -', 'e-').split()
                else:
                    _line = line.replace('\n', '')
                data.append(_line)

            shape = tuple(int(i) for i in data[0])
            i = 1
            col_headers = []
            for line in data:
                if '###' in line:
                    col_headers.append(line.split('### ')[-1])
                    if len(col_headers) == 4:
                        break

            x = []
            for n, line in enumerate(data[2:]):
                if '###' in line or len(x) >= shape[0]:
                    line_num = 3 + n
                    break
                x.extend(line)

            y = []
            for n, line in enumerate(data[line_num:]):
                if '###' in line or len(x) >= shape[1]:
                    line_num += n + 1
                    break
                y.extend(line)

            x = np.squeeze(np.array(x).astype(float))[:shape[0]]
            y = np.squeeze(np.array(y).astype(float))[:shape[1]]
            X, Y = np.meshgrid(x, y)

            _temp_int_err = []
            for i in range(shape[0] * 2):
                _temp_data = []
                for n, line in enumerate(data[line_num:]):
                    if '###' in line:
                        line_num += n + 1
                        break
                    else:
                        _temp_data.extend(line)
                _temp_int_err.append(np.squeeze(np.array(_temp_data).astype(float))[:shape[1]])
            intensity = np.array(_temp_int_err[0::2]).T
            err = np.array(_temp_int_err[1::2]).T

            self._data = OrderedDict(intensity=intensity.flatten(),
                                     error=err.flatten(),
                                     x=X.flatten(),
                                     y=X.flatten(),
                                     monitor=np.ones(len(intensity)).flatten(),
                                     time=np.ones(len(intensity)).flatten())
            self.data_keys = {'detector': 'intensity', 'monitor': 'monitor', 'time': 'time'}
            self._err = err

    def load_xyie(self, filename):
        with open(filename) as f:
            data = []
            for line in f:
                data.append(line.replace('\n', '').split())

        shape = tuple(int(i) for i in data[0])
        x = np.squeeze(np.array(data[2:2 + shape[0]]).astype(float))
        y = np.squeeze(np.array(data[3 + shape[0]:3 + shape[0] + shape[1]]).astype(float))
        i = np.array(data[4 + np.sum(shape):4 + np.sum(shape) + shape[1]]).astype(float)
        e = np.array(data[5 + np.sum(shape) + shape[1]:5 + np.sum(shape) + 2 * shape[1]]).astype(float)

        if x.shape[0] == shape[0] and y.shape[0] == shape[1] and i.T.shape == shape and e.T.shape == shape:
            X, Y = np.meshgrid(x, y)
            self._data = OrderedDict(intensity=i.flatten(),
                                     error=e.flatten(),
                                     x=X.flatten(),
                                     y=X.flatten(),
                                     monitor=np.ones(len(i)).flatten(),
                                     time=np.ones(len(i)).flatten())
            self.data_keys = {'detector': 'intensity', 'monitor': 'monitor', 'time': 'time'}
            self._err = e
        else:
            raise ValueError('File was not loaded correctly!')

    def load_xye(self, filename):
        x, y, e = np.loadtxt(filename, unpack=True)
        self._data = OrderedDict(intensity=y, error=e, x=x, monitor=np.ones(len(y)), time=np.ones(len(y)))
        self.data_keys = {'detector': 'intensity', 'monitor': 'monitor', 'time': 'time'}
        self._err = e



import matplotlib.pyplot as plt
import numpy as np


class PlotFit(object):
    """class containing fit plotting methods

    Methods
    -------
     build_param_table
     plot
    """

    def build_param_table(self,function_str=None):
       r"""Builds a table of parameters for including in a plot or for use with print

       Parameters
       ==========
       function_str : string
           A string that describes the fitting function

       """

       parms=np.array(self.params)
       d_parms=np.array(self.xerror)
       #determine number of decimal places
       # use the differnce in the number of decimal places from the parameter and the error
       # to determine how many places to keep  add one more decimal place for good measure.
       # if the difference is less than 1 make it 2.
       print(parms)
       print ("\n")
       print(d_parms)
       n_dec=np.floor(np.log10(np.abs(parms)))-np.floor(np.log10(np.abs(d_parms)))+1
       n_dec[n_dec<1]=2
       n_dec=n_dec.astype('int')
       #table is generated below here
       str_out=''
       if function_str is not None:
           str_out=str_out+function_str+'\n'
       #flabel columns
       str_out=str_out+'param\t\terr\n'
       # for each parameter append the parameter and the error keep 2 decimal places for the error to the string
       for idx in range(len(parms)):
           fmtstr=r'"%1.'+str(n_dec[idx])+'e\t\t'+r'%1.1e\n" %(parms[idx],d_parms[idx])'
           #print(fmtstr)
           str_out=str_out+eval(fmtstr)
       #append the chisqruared value
       str_out=str_out+"$\chi^2=$%f\n"%self.chi2_min
       return str_out

    def plot(self,fit_function,function_str=None,plot_residuals=True):
           r""" Plots the data and the results of a fit

           Parameters
           ----------
           fit_function :function
               the function used in the fit
           function_str :string
                A string that describes the fit function.
           residuals :boolean
                If this is True it will give a plot of the residuals

           """
           xin=self.data[1]
           yin=self.data[2]
           errin=self.data[3]
           plt.figure()
           if plot_residuals:
               ax1=plt.subplot2grid((2,3),(0,0),colspan=2)
               ax2=plt.subplot2grid((2,3),(1,0),colspan=2)
               ax2.plot(xin,self.residuals(self.params,self.data),'b-')
           else:
               ax1=plt.subplot2grid((2,3),(0,0),colspan=2,rowspan=2)
           #plot fit
           ax1.plot(xin,fit_function(self.params,xin),'b-')
           plt.hold(True)
           #plot data
           ax1.errorbar(xin,yin,yerr=errin,fmt='ro')
           # add text table
           parmtext=self.build_param_table(function_str)
           ax3=plt.subplot2grid((2,3),(0,2),rowspan=2)
           ax3.text(0.1,0.5,parmtext,size=16)
           ax3.get_xaxis().set_visible(False)
           ax3.get_yaxis().set_visible(False)
           plt.show()



# -*- coding: utf-8 -*-
import datetime as dt
from warnings import warn

import numpy as np

from ..energy import Energy
from .tools import _CleanArgs, fproject, get_bragg_widths


class PlotInstrument(object):
    r"""Class containing resolution plotting methods

    Attributes
    ----------
    description_string

    Methods
    -------
    plot_projections
    plot_ellipsoid
    plot_instrument
    plot_slice

    """

    def plot_projections(self, hkle, npts=36, dpi=100):
        r"""Plots resolution ellipses in the QxQy, QxW, and QyW zones

        Parameters
        ----------
        hkle : tup
            A tuple of intergers or arrays of H, K, L, and W (energy transfer)
            values at which resolution ellipses are desired to be plotted

        npts : int, optional
            Number of points in an individual resolution ellipse. Default: 36

        dpi : int, optional
            Specify DPI of figure. Default: 100

        """
        [H, K, L, W] = hkle

        if hasattr(self, "projections"):
            if np.all(H == self.H) and np.all(K == self.K) and np.all(L == self.L) and np.all(W == self.W):
                projections = self.projections
            else:
                self.calc_projections(hkle, npts)
                projections = self.projections
        else:
            self.calc_projections(hkle, npts)
            projections = self.projections

        import matplotlib.pyplot as plt

        plt.rc('font', **{'family': 'Bitstream Vera Sans', 'serif': 'cm10', 'size': 6})
        plt.rc('lines', markersize=3, linewidth=1)

        fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, facecolor='w', edgecolor='k', dpi=dpi)
        fig.subplots_adjust(bottom=0.175, left=0.15, right=0.85, top=0.95, wspace=0.35, hspace=0.25)

        ax1_lims, ax2_lims, ax3_lims = [[[0, 0], [0, 0]], [[0, 0], [0, 0]], [[0, 0], [0, 0]]]

        if len(self.RMS.shape) == 3:
            length = self.RMS.shape[0]
        else:
            length = 1

        for i in range(length):
            for ax, slice_str, lims in zip([ax1, ax2, ax3], ['QxQy', 'QxW', 'QyW'], [ax1_lims, ax2_lims, ax3_lims]):
                self.plot_slice(ax, slice_str, projections, self.sample.u, self.sample.v, num=i)

                xlim = ax.get_xlim()
                if xlim[0] < lims[0][0] or i == 0:
                    lims[0][0] = xlim[0]
                if xlim[1] > lims[0][1] or i == 0:
                    lims[0][1] = xlim[1]

                ylim = ax.get_ylim()
                if ylim[0] < lims[1][0] or i == 0:
                    lims[1][0] = ylim[0]
                if ylim[1] > lims[1][1] or i == 0:
                    lims[1][1] = ylim[1]

                ax.set_xlim(lims[0])
                ax.set_ylim(lims[1])

        ax4.axis('off')
        ax4.text(0, 1, self.description_string, transform=ax4.transAxes, horizontalalignment='left',
                 verticalalignment='top')

        plt.show()

    def plot_ellipsoid(self, hkle, dpi=100):
        r"""Plots the resolution ellipsoid in the $Q_x$, $Q_y$, $W$ zone

        Parameters
        ----------
        hkle : tup
            A tuple of intergers or arrays of H, K, L, and W (energy transfer)
            values at which resolution ellipsoid are desired to be plotted

        dpi : int, optional
            Number of points in the plot

        """
        try:
            from skimage import measure
            import matplotlib.pyplot as plt
            from mpl_toolkits.mplot3d import Axes3D
        except ImportError:
            raise

        # clean arguments
        [length, H, K, L, W] = _CleanArgs(*hkle)

        # check if resolution already calculated at given hkle
        try:
            if np.all(H == self.H) and np.all(K == self.K) and np.all(L == self.L) and np.all(W == self.W):
                NP = np.array(self.RMS)
                R0 = self.R0
            else:
                self.calc_resolution(hkle)
                NP = np.array(self.RMS)
                R0 = self.R0
        except AttributeError:
            self.calc_resolution(hkle)
            NP = np.array(self.RMS)
            R0 = self.R0

        if NP.shape == (4, 4):
            NP = NP[np.newaxis]
            R0 = [R0]
        else:
            NP = NP.T

        # Create a canvas with a 3D viewport
        fig = plt.figure(facecolor='w', edgecolor='k')
        ax = fig.add_subplot(111, projection='3d')

        for ind, (r0, rms) in enumerate(zip(R0, NP)):
            qx0, qy0 = [np.dot([H[ind], K[ind], L[ind]], self.orient1 / np.linalg.norm(self.orient1) ** 2),
                        np.dot([H[ind], K[ind], L[ind]], self.orient2 / np.linalg.norm(self.orient2) ** 2)]
            qw0 = W[ind]

            def resolution_ellipsoid(prefac, resmat, x, y, w, x0, y0, w0):
                r"""Resolution ellipsoid equation for qx, qy, and w dimensions

                Parameters
                ----------
                prefac : float
                resmat : ndarray(4,4)
                x : ndarray
                y : ndarray
                w : ndarray
                x0 : float
                y0 : float
                w0 : float

                Returns
                -------
                    res_func : ndarray
                """
                ee = (resmat[0, 0] * (x - x0) ** 2 +
                      resmat[1, 1] * (y - qy0) ** 2 +
                      resmat[3, 3] * (w - w0) ** 2 +
                      2 * resmat[0, 1] * (x - x0) * (y - y0) +
                      2 * resmat[0, 2] * (x - x0) * (w - w0) +
                      2 * resmat[1, 2] * (y - y0) * (w - w0))

                return prefac * np.exp(-1 / 2 * ee)

            # get fwhm to generate grid for plotting
            w = []
            for i in range(3):
                _rms = np.delete(np.delete(rms, 2, axis=0), 2, axis=1)
                w.append(fproject(_rms.reshape((3, 3, 1)), i)[0] * 1.01)
            wx, wy, ww = w

            # build grid
            xg, yg, zg = np.mgrid[qx0 - wx:qx0 + wx:(dpi + 1) * 1j,
                                  qy0 - wy:qy0 + wy:(dpi + 1) * 1j,
                                  qw0 - ww:qw0 + ww:(dpi + 1) * 1j]

            # resolution function (ellipsoidal gaussian)
            vol = resolution_ellipsoid(r0, rms, xg, yg, zg, qx0, qy0, qw0)

            # isosurface of resolution function at half-max
            vertices, faces, normals, values = measure.marching_cubes(vol, vol.max() / 2.0, spacing=(wx / dpi, wy / dpi, ww / dpi))

            # properly center ellipsoid
            x_verts = vertices[:, 0] + qx0 - wx / 2
            y_verts = vertices[:, 1] + qy0 - wy / 2
            z_verts = vertices[:, 2] + qw0 - ww / 2

            # plot ellipsoid
            ax.plot_trisurf(x_verts, y_verts, faces, z_verts, lw=0, cmap='Spectral')

        # figure labels
        ax.ticklabel_format(style='plain', useOffset=False)
        ax.set_xlabel(r'$q_x$ (along {0}) (r.l.u.)'.format(self.orient1), fontsize=12)
        ax.set_ylabel(r'$q_y$ (along {0}) (r.l.u.)'.format(self.orient2), fontsize=12)
        ax.set_zlabel(r'$\hbar \omega$ (meV)', fontsize=12)

        plt.show()

    def plot_instrument(self, hkle):
        r"""Plots the instrument configuration using angles for a given position
        in Q and energy transfer

        Parameters
        ----------
        hkle : tup
            A tuple of intergers or arrays of H, K, L, and W (energy transfer)
            values at which the instrument setup should be plotted

        """
        try:
            import matplotlib.pyplot as plt
            from mpl_toolkits.mplot3d import Axes3D
        except ImportError:
            raise

        fig = plt.figure(edgecolor='k', facecolor='w', figsize=plt.figaspect(0.4) * 1.25)
        ax = fig.gca(projection='3d')

        if getattr(self, "instrument_type") == "tas":
            measurements = {'guide.width': 1, 'guide.height': 1, 'mono.width': 1, 'mono.height': 1, 'sample.width': 1,
                            'sample.height': 1, 'sample.depth': 1, 'ana.width': 1, 'ana.height': 1, 'detector.width': 1,
                            'detector.height': 1, 'arms': [10, 10, 10, 10]}

            for key, value in measurements.items():
                if '.' in key:
                    if hasattr(getattr(self, key.split('.')[0]), key.split('.')[1]):
                        measurements[key] = getattr(getattr(self, key.split('.')[0]), key.split('.')[1])
                else:
                    if hasattr(self, key):
                        measurements[key] = getattr(self, key)

            angles = self.get_angles_and_Q(hkle)[0]
            distances = measurements['arms']

            angles = -np.deg2rad(angles)
            x, y, direction = np.zeros(3)

            # plot the Source -----------------------------------------------------
            translate = 0
            rotate = 0 * (np.pi / 180)
            direction += rotate
            x += translate * np.sin(direction)
            y += translate * np.cos(direction)

            # create a square source
            X = np.array(
                [-measurements['guide.width'] / 2, -measurements['guide.width'] / 2, measurements['guide.width'] / 2,
                measurements['guide.width'] / 2, -measurements['guide.width'] / 2])
            Z = np.array(
                [measurements['guide.height'] / 2, -measurements['guide.height'] / 2, -measurements['guide.height'] / 2,
                measurements['guide.height'] / 2, measurements['guide.height'] / 2])
            Y = np.zeros(5)
            l = ax.plot(X + x, Y + y, zs=Z, color='b')
            t = ax.text(X[0] + x, Y[0] + y, Z[0], 'Beam/Source', color='b')

            x0, y0 = x, y

            # plot the Monochromator ----------------------------------------------
            translate = distances[0]
            rotate = 0
            direction += rotate
            x += translate * np.sin(direction)
            y += translate * np.cos(direction)

            l = ax.plot([x, x0], [y, y0], zs=[0, 0], color='cyan', linestyle='--')

            # create a square Monochromator
            X = np.array([-measurements['mono.width'] / 2, -measurements['mono.width'] / 2, measurements['mono.width'] / 2,
                        measurements['mono.width'] / 2, -measurements['mono.width'] / 2]) * np.sin(angles[0])
            Z = np.array(
                [measurements['mono.height'] / 2, -measurements['mono.height'] / 2, -measurements['mono.height'] / 2,
                measurements['mono.height'] / 2, measurements['mono.height'] / 2])
            Y = X * np.cos(angles[0])
            l = ax.plot(X + x, Y + y, zs=Z, color='r')
            t = ax.text(X[0] + x, Y[0] + y, Z[0], 'Monochromator', color='r')

            x0, y0 = x, y

            # plot the Sample -----------------------------------------------------
            translate = distances[1]

            rotate = angles[1]
            direction += rotate
            x += translate * np.sin(direction)
            y += translate * np.cos(direction)

            l = ax.plot([x, x0], [y, y0], zs=[0, 0], color='cyan', linestyle='--')

            # create a rotated square Sample
            X = np.array(
                [-measurements['sample.width'] / 2, -measurements['sample.width'] / 2, measurements['sample.width'] / 2,
                measurements['sample.width'] / 2, -measurements['sample.width'] / 2]) * np.sin(angles[2])
            Z = np.array(
                [measurements['sample.height'] / 2, -measurements['sample.height'] / 2, -measurements['sample.height'] / 2,
                measurements['sample.height'] / 2, measurements['sample.height'] / 2])
            Y = X * np.cos(angles[2])
            l1 = ax.plot(X + x, Y + y, zs=Z, color='g')
            t = ax.text(X[0] + x, Y[0] + y, Z[0], 'Sample', color='g')
            X = np.array(
                [-measurements['sample.depth'] / 2, -measurements['sample.depth'] / 2, measurements['sample.depth'] / 2,
                measurements['sample.depth'] / 2, -measurements['sample.depth'] / 2]) * np.sin(
                angles[2] + np.pi / 2)
            Z = np.array(
                [measurements['sample.height'] / 2, -measurements['sample.height'] / 2, -measurements['sample.height'] / 2,
                measurements['sample.height'] / 2, measurements['sample.height'] / 2])
            Y = X * np.cos(angles[2] + np.pi / 2)
            l2 = ax.plot(X + x, Y + y, zs=Z, color='g')

            x0, y0 = x, y

            # plot the Analyzer ---------------------------------------------------
            translate = distances[2]

            rotate = angles[3]
            direction += rotate

            x += translate * np.sin(direction)
            y += translate * np.cos(direction)
            l = ax.plot([x, x0], [y, y0], zs=[0, 0], color='cyan', linestyle='--')

            # create a square
            X = np.array([-measurements['ana.width'] / 2, -measurements['ana.width'] / 2, measurements['ana.width'] / 2,
                        measurements['ana.width'] / 2, -measurements['ana.width'] / 2]) * np.sin(angles[4])
            Z = np.array([measurements['ana.height'] / 2, -measurements['ana.height'] / 2, -measurements['ana.height'] / 2,
                        measurements['ana.height'] / 2, measurements['ana.height'] / 2])
            Y = X * np.cos(angles[4])
            l = ax.plot(X + x, Y + y, zs=Z, color='magenta')
            t = ax.text(X[0] + x, Y[0] + y, Z[0], 'Analyzer', color='magenta')

            x0, y0 = x, y

            # plot the Detector ---------------------------------------------------
            translate = distances[3]
            rotate = angles[5]
            direction += rotate

            x += translate * np.sin(direction)
            y += translate * np.cos(direction)
            l = ax.plot([x, x0], [y, y0], zs=[0, 0], color='cyan', linestyle='--')

            # create a square
            X = np.array(
                [-measurements['detector.width'] / 2, -measurements['detector.width'] / 2,
                measurements['detector.width'] / 2, measurements['detector.width'] / 2,
                -measurements['detector.width'] / 2])
            Z = np.array(
                [measurements['detector.height'] / 2, -measurements['detector.height'] / 2,
                -measurements['detector.height'] / 2, measurements['detector.height'] / 2,
                measurements['detector.height'] / 2])
            Y = np.zeros(5)
            l = ax.plot(X + x, Y + y, zs=Z, color='k')
            t = ax.text(X[0] + x, Y[0] + y, Z[0], 'Detector', color='k')

            ax.set_zlim3d(getattr(ax, 'get_zlim')()[0], getattr(ax, 'get_zlim')()[1] * 10)

        elif getattr(self, "instrument_type") == "tof":
            warn("Plotting of Time of Flight Instrument layout is not yet supported")

        plt.show()

    @staticmethod
    def plot_slice(axis, qslice, projections, u, v, num=0):

        r"""Class method for plotting individual projections. Plots both
        slices and projections.

        Parameters
        ----------
        axis : matplotlib_axis
            Axis to which to plot the selected projection

        qslice : str
            The projection to plot. 'QxQy', 'QxW' or 'QyW'.

        projections : dict
            Projections as generated by :py:meth:`.Instrument.calc_projections`

        u : ndarray(3)
            First orientation vector.

        v : ndarray(3)
            Second orientation vector.

        num : int, optional
            If multiple projections are present, choose which one to plot

        """
        dQ1, dQ2 = [], []

        axis.fill(projections[qslice][num, 0, :], projections[qslice][num, 1, :], zorder=0, alpha=0.5,
                  edgecolor='none')
        axis.plot(projections[qslice + 'Slice'][num, 0, :], projections[qslice + 'Slice'][num, 1, :], zorder=1)

        dQ1.append(np.max(projections[qslice][num, 0, :]) - np.min(projections[qslice][num, 0, :]))
        dQ2.append(np.max(projections[qslice][num, 1, :]) - np.min(projections[qslice][num, 1, :]))

        axis.set_xlim(np.min(projections[qslice][num, 0, :]), np.max(projections[qslice][num, 0, :]))
        axis.set_ylim(np.min(projections[qslice][num, 1, :]), np.max(projections[qslice][num, 1, :]))

        dQ1, dQ2 = [np.max(item) for item in [dQ1, dQ2]]

        axis.set_xlabel(r'$\mathbf{Q}_1$ (along ' + str(u) + r') (r.l.u.)' + r', $\delta Q_1={0:.3f}$'.format(dQ1),
                        fontsize=7)
        if 'W' in qslice:
            axis.set_ylabel(r'$\hbar \omega$ (meV)' + r', $\delta \hbar \omega={0:.3f}$'.format(dQ2), fontsize=7)
        else:
            axis.set_ylabel(r'$\mathbf{Q}_2$ (along ' + str(v) + r') (r.l.u.)' + r', $\delta Q_2={0:.3f}$'.format(dQ2),
                            fontsize=7)

    @property
    def description_string(self):
        r"""Generates text string describing most recent resolution calculation

        """
        try:
            method = ['Cooper-Nathans', 'Popovici'][self.method]
        except AttributeError:
            method = 'Cooper-Nathans'
        frame = '[Q1,Q2,Qz,E]'

        if hasattr(self, 'infin'):
            FX = 2 * int(self.infin == -1) + int(self.infin == 1)
        else:
            FX = 2

        if self.RMS.shape == (4, 4):
            NP = self.RMS
            R0 = float(self.R0)
            hkle = self.HKLE
        else:
            NP = self.RMS[0]
            R0 = self.R0[0]
            hkle = [self.H[0], self.K[0], self.L[0], self.W[0]]

        ResVol = (2 * np.pi) ** 2 / np.sqrt(np.linalg.det(NP))
        bragg_widths = get_bragg_widths(NP)

        if getattr(self, "instrument_type") == "tas":
            angles = self.get_angles_and_Q(hkle)[0]

            text_format = ['Method: {0}'.format(method),
                           'Position HKLE [{0}]\n'.format(dt.datetime.now().strftime('%d-%b-%Y %H:%M:%S')),
                           ' [Q_H, Q_K, Q_L, E] = {0} \n'.format(self.HKLE),
                           'Resolution Matrix M in {0} (M/10^4):'.format(frame),
                           '[[{0:.4f}   {1:.4f}   {2:.4f}   {3:.4f}]'.format(*NP[:, 0] / 1.0e4),
                           ' [{0:.4f}   {1:.4f}   {2:.4f}   {3:.4f}]'.format(*NP[:, 1] / 1.0e4),
                           ' [{0:.4f}   {1:.4f}   {2:.4f}   {3:.4f}]'.format(*NP[:, 2] / 1.0e4),
                           ' [{0:.4f}   {1:.4f}   {2:.4f}   {3:.4f}]]\n'.format(*NP[:, 3] / 1.0e4),
                           'Resolution volume:   V_0={0:.6f} meV/A^3'.format(2 * ResVol),
                           'Intensity prefactor: R_0={0:.3f}'.format(R0),
                           'Bragg width in [Q_1,Q_2,E] (FWHM):',
                           ' dQ_1={0:.3f} dQ_2={1:.3f} [A-1] dE={2:.3f} [meV]'.format(bragg_widths[0], bragg_widths[1],
                                                                                       bragg_widths[4]),
                           ' dQ_z={0:.3f} Vanadium width V={1:.3f} [meV]'.format(*bragg_widths[2:4]),
                           'Instrument parameters:',
                           ' DM  =  {0:.3f} ETAM= {1:.3f} SM={2}'.format(self.mono.d, self.mono.mosaic, self.mono.dir),
                           ' KFIX=  {0:.3f} FX  = {1} SS={2}'.format(Energy(energy=self.efixed).wavevector, FX,
                                                                       self.sample.dir),
                           ' DA  =  {0:.3f} ETAA= {1:.3f} SA={2}'.format(self.ana.d, self.ana.mosaic, self.ana.dir),
                           ' A1= {0:.2f} A2={1:.2f} A3={2:.2f} A4={3:.2f} A5={4:.2f} A6={5:.2f} [deg]'.format(*angles),
                           'Collimation [arcmin]:',
                           ' Horizontal: [{0:.0f}, {1:.0f}, {2:.0f}, {3:.0f}]'.format(*self.hcol),
                           ' Vertical: [{0:.0f}, {1:.0f}, {2:.0f}, {3:.0f}]'.format(*self.vcol),
                           'Sample:',
                           ' a, b, c  =  [{0}, {1}, {2}] [Angs]'.format(self.sample.a, self.sample.b, self.sample.c),
                           ' Alpha, Beta, Gamma  =  [{0}, {1}, {2}] [deg]'.format(self.sample.alpha, self.sample.beta,
                                                                                  self.sample.gamma),
                           ' U  =  {0} [rlu]   V  =  {1} [rlu]'.format(self.orient1, self.orient2)]

        elif getattr(self, "instrument_type") == "tof":
            Q = self.HKLE[:3]
            theta = self.sample.get_two_theta(Q, self.ei.wavelength)
            phi = np.rad2deg(np.pi / 2.0 - np.deg2rad(self.sample.get_phi(Q)))

            tau_p = getattr(self, "tau_p", self.choppers[0].tau)
            tau_m = getattr(self, "tau_m", self.choppers[1].tau)
            tau_d = getattr(self, "tau_d", self.detector.tau)

            l_pm = getattr(self, "l_pm", np.abs(np.subtract(self.choppers[0].distance, self.choppers[1].distance)))
            l_ms = getattr(self, "l_ms", self.sample.distance)
            l_sd = getattr(self, "l_sd", self.detector.radius)

            sigma_l_pm = getattr(self, "sigma_l_pm", self.guides[0].sigma_l)
            sigma_l_ms = getattr(self, "sigma_l_ms", self.guides[1].sigma_l)
            sigma_l_sd = getattr(self, "sigma_l_sd", self.detector.radius / np.cos(np.deg2rad(getattr(self.sample, "mosaic", 60) / 60)) - self.detector.radius)

            sigma_theta_i = getattr(self, "sigma_theta_i", self.guides[1].sigma_theta * self.ei.wavelength)
            sigma_phi_i = getattr(self, "sigma_phi_i", self.guides[1].sigma_phi * self.ei.wavelength)

            sigma_theta = getattr(self, "sigma_theta", np.max([getattr(self.sample, "mosaic", 60.0) / 60.0, self.detector.hpixels]))
            sigma_phi = getattr(self, "sigma_phi", np.max([getattr(self.sample, "vmosaic", 60.0) / 60.0, np.max(self.detector.height) / self.detector.vpixels]))

            theta_i = getattr(self, "theta_i", 0.0)
            phi_i = getattr(self, "phi_i", 0.0)

            text_format = ["Method: Violini",
                           'Position HKLE [{0}]\n'.format(dt.datetime.now().strftime('%d-%b-%Y %H:%M:%S')),
                           ' [Q_H, Q_K, Q_L, E] = {0} \n'.format(self.HKLE),
                           'Resolution Matrix M in {0} (M/10^4):'.format(frame),
                           '[[{0:.4f}   {1:.4f}   {2:.4f}   {3:.4f}]'.format(*NP[:, 0] / 1.0e4),
                           ' [{0:.4f}   {1:.4f}   {2:.4f}   {3:.4f}]'.format(*NP[:, 1] / 1.0e4),
                           ' [{0:.4f}   {1:.4f}   {2:.4f}   {3:.4f}]'.format(*NP[:, 2] / 1.0e4),
                           ' [{0:.4f}   {1:.4f}   {2:.4f}   {3:.4f}]]\n'.format(*NP[:, 3] / 1.0e4),
                           'Resolution volume:   V_0={0:.6f} meV/A^3'.format(2 * ResVol),
                           'Intensity prefactor: R_0={0:.3f}'.format(R0),
                           'Bragg width in [Q_1,Q_2,E] (FWHM):',
                           ' dQ_1={0:.3f} dQ_2={1:.3f} [A-1] dE={2:.3f} [meV]'.format(bragg_widths[0], bragg_widths[1],
                                                                                       bragg_widths[4]),
                           ' dQ_z={0:.3f} Vanadium width V={1:.3f} [meV]'.format(*bragg_widths[2:4]),
                           'Instrument parameters:',
                           ' taup =  {0:.1f} taum = {1:.1f} tamd = {2:.1f} [us]'.format(tau_p, tau_m, tau_d),
                           ' l_pm =  {0:.1f} l_ms = {1:.1f} l_sd = {2:.1f} [cm]'.format(l_pm, l_ms, l_sd),
                           ' sig_l_pm=  {0:.2f} sig_l_ms= {1:.2f} sig_l_sd= {2:.2f} [cm]'.format(sigma_l_pm, sigma_l_ms,
                                                                                                 sigma_l_sd),
                           ' theta_i = {0:.3f} phi_i = {1:.3f} [deg]'.format(theta_i, phi_i),
                           " sig_theta_i = {0:.3f} sig_phi_i = {1:.3f} [deg]".format(sigma_theta_i, sigma_phi_i),
                           " 2theta = {0:.3f} phi = {1:.3f} [deg]".format(theta, phi),
                           " sig_theta = {0:.3f} sig_phi = {1:.3f} [deg]".format(sigma_theta, sigma_phi),
                           'Sample:',
                           ' a, b, c  =  [{0}, {1}, {2}] [angstrom]'.format(self.sample.a, self.sample.b, self.sample.c),
                           ' alpha, beta, gamma  =  [{0}, {1}, {2}] [deg]'.format(self.sample.alpha, self.sample.beta,
                                                                                  self.sample.gamma),
                           ' U  =  {0} [rlu]   V  =  {1} [rlu]'.format(self.orient1, self.orient2)]

        return '\n'.join(text_format)


class PlotTofInstrument(object):
    def __init__(self):
        pass



# -*- coding: utf-8 -*-
r"""Handles lattice geometries to find rotations and transformations

"""
import numpy as np


class Lattice(object):
    u"""Class to describe a generic lattice system defined by lattice six
    parameters, (three constants and three angles).

    Parameters
    ----------
    a : float
        Lattice constant *a* in \u212B

    b : float
        Lattice constant *b* in \u212B

    c : float
        Lattice constant *c* in \u212B

    alpha : float
        Lattice angle \U0001D6FC in degrees

    beta : float
        Lattice angle \U0001D6FD in degrees

    gamma : float
        Lattice angle \U0001D6FE in degrees

    Returns
    -------
    lattice : object
        Object containing lattice information

    Attributes
    ----------
    a
    b
    c
    astar
    bstar
    cstar
    alpha
    beta
    gamma
    alpha_rad
    beta_rad
    gamma_rad
    alphastar
    betastar
    gammastar
    alphastar_rad
    betastar_rad
    gammastar_rad
    abg_rad
    reciprocal_abc
    reciprocal_abg
    reciprocal_abg_rad
    lattice_type
    volume
    reciprocal_volume
    G
    Gstar
    Bmatrix

    Methods
    -------
    get_d_spacing
    get_q
    get_two_theta
    get_angle_between_planes

    """

    def __init__(self, a, b, c, alpha, beta, gamma):
        self.abc = [a, b, c]
        self.abg = [alpha, beta, gamma]

    def __repr__(self):
        return "Lattice({0}, {1}, {2}, {3}, {4}, {5})".format(self.a, self.b, self.c, self.alpha, self.beta, self.gamma)

    @property
    def a(self):
        r"""First lattice constant in Angstrom
        """
        return self.abc[0]

    @a.setter
    def a(self, a):
        self.abc[0] = a

    @property
    def b(self):
        r"""Second lattice constant in Angstrom

        """
        return self.abc[1]

    @b.setter
    def b(self, b):
        self.abc[1] = b

    @property
    def c(self):
        r"""Third lattice constant in Angstrom
        """
        return self.abc[2]

    @c.setter
    def c(self, c):
        self.abc[2] = c

    @property
    def alpha(self):
        r"""First lattice angle in degrees
        """
        return self.abg[0]

    @alpha.setter
    def alpha(self, alpha):
        self.abg[0] = alpha

    @property
    def beta(self):
        r"""Second lattice angle in degrees
        """
        return self.abg[1]

    @beta.setter
    def beta(self, beta):
        self.abg[1] = beta

    @property
    def gamma(self):
        r"""Third lattice angle in degrees
        """
        return self.abg[2]

    @gamma.setter
    def gamma(self, gamma):
        self.abg[2] = gamma

    @property
    def alpha_rad(self):
        r"""First lattice angle in radians
        """
        return self.abg_rad[0]

    @property
    def beta_rad(self):
        r"""Second lattice angle in radians
        """
        return self.abg_rad[1]

    @property
    def gamma_rad(self):
        r"""Third lattice angle in radians
        """
        return self.abg_rad[2]

    @property
    def astar(self):
        r"""First inverse lattice constant in inverse Angstrom
        """
        return self.b * self.c * np.sin(self.alpha_rad) / self.volume * 2 * np.pi

    @property
    def bstar(self):
        r"""Second inverse lattice constant in inverse Angstrom
        """
        return self.a * self.c * np.sin(self.beta_rad) / self.volume * 2 * np.pi

    @property
    def cstar(self):
        r"""Third inverse lattice constant in inverse Angstrom
        """
        return self.a * self.b * np.sin(self.gamma_rad) / self.volume * 2 * np.pi

    @property
    def alphastar_rad(self):
        r"""First inverse lattice angle in radians
        """
        return np.arccos((np.cos(self.beta_rad) * np.cos(self.gamma_rad) -
                          np.cos(self.alpha_rad)) /
                         (np.sin(self.beta_rad) * np.sin(self.gamma_rad)))

    @property
    def betastar_rad(self):
        r"""Second inverse lattice angle in radians
        """
        return np.arccos((np.cos(self.alpha_rad) *
                          np.cos(self.gamma_rad) -
                          np.cos(self.beta_rad)) /
                         (np.sin(self.alpha_rad) * np.sin(self.gamma_rad)))

    @property
    def gammastar_rad(self):
        r"""Third inverse lattice angle in radians
        """
        return np.arccos((np.cos(self.alpha_rad) * np.cos(self.beta_rad) -
                          np.cos(self.gamma_rad)) /
                         (np.sin(self.alpha_rad) * np.sin(self.beta_rad)))

    @property
    def alphastar(self):
        r"""First inverse lattice angle in degrees
        """
        return np.rad2deg(self.alphastar_rad)

    @property
    def betastar(self):
        r"""First inverse lattice angle in degrees
        """
        return np.rad2deg(self.betastar_rad)

    @property
    def gammastar(self):
        r"""First inverse lattice angle in degrees
        """
        return np.rad2deg(self.gammastar_rad)

    @property
    def reciprocal_abc(self):
        r"""Reciprocal lattice constants in inverse Angstrom returned in list
        """
        return [self.astar, self.bstar, self.cstar]

    @property
    def reciprocal_abg(self):
        r"""Reciprocal lattice angles in degrees returned in list
        """
        return [self.alphastar, self.betastar, self.gammastar]

    @property
    def reciprocal_abg_rad(self):
        r"""Reciprocal lattice angles in radians returned in list
        """
        return [self.alphastar_rad, self.betastar_rad, self.gammastar_rad]

    @property
    def abg_rad(self):
        r"""Lattice angles in radians returned in list
        """
        return np.deg2rad(self.abg)

    @property
    def lattice_type(self):
        r"""Type of lattice determined by the provided lattice constants and angles

        """

        if len(np.unique(self.abc)) == 3 and len(np.unique(self.abg)) == 3:
            return 'triclinic'
        elif len(np.unique(self.abc)) == 3 and self.abg[1] != 90 and np.all(np.array(self.abg)[:3:2] == 90):
            return 'monoclinic'
        elif len(np.unique(self.abc)) == 3 and np.all(np.array(self.abg) == 90):
            return 'orthorhombic'
        elif len(np.unique(self.abc)) == 1 and len(np.unique(self.abg)) == 1 and np.all(
                        np.array(self.abg) < 120) and np.all(np.array(self.abg) != 90):
            return 'rhombohedral'
        elif len(np.unique(self.abc)) == 2 and self.abc[0] == self.abc[1] and np.all(np.array(self.abg) == 90):
            return 'tetragonal'
        elif len(np.unique(self.abc)) == 2 and self.abc[0] == self.abc[1] and np.all(np.array(self.abg)[0:2] == 90) and \
                        self.abg[2] == 120:
            return 'hexagonal'
        elif len(np.unique(self.abc)) == 1 and np.all(np.array(self.abg) == 90):
            return 'cubic'
        else:
            raise ValueError('Provided lattice constants and angles do not resolve to a valid Bravais lattice')

    @property
    def volume(self):
        u"""Volume of the unit cell in \u212B\ :sup:`3`

        """
        return np.sqrt(np.linalg.det(self.G))

    @property
    def reciprocal_volume(self):
        u"""Volume of the reciprocal unit cell in (\u212B\ :sup:`-1`\ )\ :sup:`3`

        """
        return np.sqrt(np.linalg.det(self.Gstar))

    @property
    def G(self):
        r"""Metric tensor of the real space lattice

        """

        a, b, c = self.abc
        alpha, beta, gamma = self.abg_rad

        return np.matrix([[a ** 2, a * b * np.cos(gamma), a * c * np.cos(beta)],
                          [a * b * np.cos(gamma), b ** 2, b * c * np.cos(alpha)],
                          [a * c * np.cos(beta), b * c * np.cos(alpha), c ** 2]], dtype=float)

    @property
    def Gstar(self):
        r"""Metric tensor of the reciprocal lattice

        """

        return np.linalg.inv(self.G) * 4 * np.pi ** 2

    @property
    def Bmatrix(self):
        r"""Cartesian basis matrix in reciprocal units such that
        Bmatrix*Bmatrix.T = Gstar

        """

        return np.matrix([[self.astar, self.bstar * np.cos(self.gammastar_rad),  self.cstar * np.cos(self.betastar_rad)],
                          [0, self.bstar * np.sin(self.gammastar_rad), -self.cstar * np.sin(self.betastar_rad) * np.cos(self.alpha_rad)],
                          [0, 0, self.cstar * np.sin(self.betastar_rad) * np.sin(self.alphastar_rad)]], dtype=float)

    def get_d_spacing(self, hkl):
        u"""Returns the d-spacing of a given reciprocal lattice vector.

        Parameters
        ----------
        hkl : array_like
            Reciprocal lattice vector in r.l.u.

        Returns
        -------
        d : float
            The d-spacing in \u212B

        """
        hkl = np.array(hkl)

        return float(1 / np.sqrt(np.dot(np.dot(hkl, self.Gstar / 4 / np.pi ** 2), hkl)))

    def get_angle_between_planes(self, v1, v2):
        r"""Returns the angle :math:`\phi` between two reciprocal lattice
        vectors (or planes as defined by the vectors normal to the plane).

        Parameters
        ----------
        v1 : array_like
            First reciprocal lattice vector in units r.l.u.

        v2 : array_like
            Second reciprocal lattice vector in units r.l.u.

        Returns
        -------
        phi : float
            The angle between v1 and v2 in degrees

        """

        v1, v2 = np.array(v1), np.array(v2)

        return float(np.rad2deg(np.arccos(np.inner(np.inner(v1, self.Gstar), v2) /
                                          np.sqrt(np.inner(np.inner(v1, self.Gstar), v1)) /
                                          np.sqrt(np.inner(np.inner(v2, self.Gstar), v2)))))

    def get_two_theta(self, hkl, wavelength):
        u"""Returns the detector angle 2\U0001D703 for a given reciprocal
        lattice vector and incident wavelength.

        Parameters
        ----------
        hkl : array_like
            Reciprocal lattice vector in r.l.u.

        wavelength : float
            Wavelength of the incident beam in \u212B

        Returns
        -------
        two_theta : float
            The angle of the detector 2\U0001D703 in degrees

        """

        return 2 * np.rad2deg(np.arcsin(wavelength / 2 /
                                        self.get_d_spacing(hkl)))

    def get_q(self, hkl):
        u"""Returns the magnitude of *Q* for a given reciprocal lattice
        vector in \u212B\ :sup:`-1`.

        Parameters
        ----------
        hkl : array_like
            Reciprocal lattice vector in r.l.u.

        Returns
        -------
        q : float
            The magnitude of the reciprocal lattice vector *Q* in
            \u212B\ :sup:`-1`

        """

        return 2 * np.pi / self.get_d_spacing(hkl)



# -*- coding: utf-8 -*-
r"""Define a Time Of Flight instrument for resolution calculations

"""
import numpy as np

from ..constants import e, hbar, neutron_mass
from ..crystal.sample import Sample
from ..energy import Energy
from .chopper import Chopper
from .detector import Detector
from .exceptions import DetectorError
from .general import GeneralInstrument
from .guide import Guide
from .plot import PlotInstrument
from .tools import _CleanArgs, chop, get_angle_ki_Q, get_kfree


class TimeOfFlightInstrument(GeneralInstrument, PlotInstrument):
    r"""An object representing a Time of Flight (TOF) instrument experimental
    configuration, including a sample.

    Parameters
    ----------
    ei : float, optional
        Incident energy in meV. Default: 3.0

    choppers : obj or list, optional
        Chopper object or list of chopper objects defining all choppers.
        If list, choppers will be automatically sorted by distance from
        source. Default: See code.

    sample : obj, optional
        Sample objection defining the properties of the sample. Default: See code.

    detector : obj, optional
        Detector object defining the properties of the detector. Default: See code.

    guides : obj, optional

    theta_i : float, optional
        Incident 2-theta in degrees. Default: 0

    phi_i : float, optional
        Incident phi in degrees. Default: 0

    Notes
    -----
    If no initial values are specified, the default instrument is similar to
    the T-REX concept for the ESS specified in Violini et al. (2014).

    Calculations of the resolution are based on equations in N. Violini et al.,
    Nuclear Instruments and Methods in Physics Research A 736 (2014) 31-39.

    Attributes
    ----------
    ei
    choppers
    sample
    detector
    guides
    description_string

    Methods
    -------
    calc_resolution
    calc_resolution_in_Q_coords
    calc_projections
    get_resolution_params
    get_resolution
    plot_projections
    plot_ellipsoid
    plot_instrument
    plot_slice

    """

    def __init__(self, ei=3.0, choppers=None, sample=None, detector=None, guides=None, theta_i=0, phi_i=0, **kwargs):
        self._ei = Energy(energy=ei)

        if choppers:
            if isinstance(choppers, list):
                self.choppers = sorted(choppers, key=lambda x: x.distance)
            else:
                self.choppers = [choppers]
        else:
            p_chopper = Chopper(0.0, 60.0, 5.0, 10, 'disk', tau=38)
            m_chopper = Chopper(2500.0, 240.0, 2.0, 10, 'disk', 2, tau=6)
            self.choppers = [p_chopper, m_chopper]

        if sample:
            self.sample = sample
        else:
            self.sample = Sample(a=6, b=7, c=8, alpha=90, beta=90, gamma=90, u=[1, 0, 0], v=[0, 1, 0], distance=3600.0)

        if detector:
            self.detector = detector
        else:
            self.detector = Detector('cylindrical', [-90, 90], [-15, 15], 300.0, 0.1, 256)

        if guides:
            self.guides = guides
        else:
            pm_guide = Guide(1, self.choppers[1].distance - self.choppers[0].distance, 15, 15)
            ms_guide = Guide(1, self.sample.distance - self.choppers[1].distance, 15, 15)
            self.guides = [pm_guide, ms_guide]

        self.theta_i = theta_i
        self.phi_i = phi_i

        for kwarg, value in kwargs.items():
            setattr(self, kwarg, value)

    def __repr__(self):
        return "Instrument('tof', engine='neutronpy', ei={0})".format(self.ei)

    @property
    def ei(self):
        r"""Incident Energy object of type :py:class:`.Energy`

        """
        return self._ei

    @ei.setter
    def ei(self, value):
        self._ei = Energy(energy=value)

    @property
    def orient1(self):
        return self.sample.u

    @property
    def orient2(self):
        return self.sample.v

    def calc_resolution_in_Q_coords(self, Q, W):
        r"""For a momentum transfer Q and energy transfers W, given experimental
        conditions specified in EXP, calculates the Cooper-Nathans or Popovici
        resolution matrix RM and resolution prefactor R0 in the Q coordinate
        system (defined by the scattering vector and the scattering plane,
        [Q_prep, Q_para, Qz, W]).

        Parameters
        ----------
        Q : ndarray or list of ndarray
            The Q vectors in reciprocal space at which resolution should be
            calculated, in rlu

        W : float or list of floats
            The energy transfers at which resolution should be calculated in meV

        Returns
        -------
        [R0, RM] : list(float, ndarray)
            Resolution pre-factor (R0) and resolution matrix (RM) at the given
            reciprocal lattice vectors and energy transfers

        Notes
        -----
            Translated from "A method to compute the covariance matrix of
            wavevector-energy transfer for neutron time-of-flight spectrometers"
            by Violini et al., Nuclear Instruments and Methods in Physics Research
            A 736 (2014) 31-39. DOI: 10.1016/j.nima.2013.10.042

        """
        # Definitions
        tau_p = getattr(self, "tau_p", self.choppers[0].tau) / 1e6
        tau_m = getattr(self, "tau_m", self.choppers[1].tau) / 1e6
        tau_d = getattr(self, "tau_d", self.detector.tau) / 1e6

        l_pm = getattr(self, "l_pm", np.abs(np.subtract(self.choppers[0].distance, self.choppers[1].distance))) / 1e2
        l_ms = getattr(self, "l_ms", self.sample.distance) / 1e2
        l_sd = getattr(self, "l_sd", self.detector.radius) / 1e2

        sigma_l_pm = getattr(self, "sigma_l_pm", self.guides[0].sigma_l) / 1e2
        sigma_l_ms = getattr(self, "sigma_l_ms", self.guides[1].sigma_l) / 1e2
        sigma_l_sd = getattr(self, "sigma_l_sd", self.detector.radius / np.cos(np.deg2rad(getattr(self.sample, "mosaic", 60) / 60)) - self.detector.radius) / 1e2

        sigma_theta_i = np.deg2rad(getattr(self, "sigma_theta_i", self.guides[1].sigma_theta * self.ei.wavelength))
        sigma_phi_i = np.deg2rad(getattr(self, "sigma_phi_i", self.guides[1].sigma_phi * self.ei.wavelength))

        sigma_theta = np.deg2rad(getattr(self, "sigma_theta", np.max([getattr(self.sample, "mosaic", 60.0) / 60.0, self.detector.hpixels])))
        sigma_phi = np.deg2rad(getattr(self, "sigma_phi", np.max([getattr(self.sample, "vmosaic", 60.0) / 60.0, np.max(self.detector.height) / self.detector.vpixels])))

        theta_i = np.deg2rad(getattr(self, "theta_i", 0.0))
        phi_i = np.deg2rad(getattr(self, "phi_i", 0.0))

        # Get ki, kf and related from ei and energy transfer
        ki = self.ei.wavevector
        kf = get_kfree(W, ki)

        vi = self.ei.velocity
        vf = Energy(wavevector=kf).velocity

        ti = l_pm / vi
        tf = l_sd / vf

        # Eq. 20 & 21 Violini :: time sigmas
        sigma_t = np.sqrt(tau_p ** 2 + tau_m ** 2)
        sigma_t_md = np.sqrt(tau_m ** 2 + tau_d ** 2)

        # Get TwoTheta and Phi from Q
        theta = np.deg2rad(self.sample.get_two_theta(Q, self.ei.wavelength))
        phi = np.pi / 2.0 - np.deg2rad(self.sample.get_phi(Q))

        # neutron mass in kg
        m_n = neutron_mass * 1e-3

        # Velocity Vector
        vel = np.matrix([[vi, vf]]) * m_n / hbar

        # Appendix A.1 Violini :: spherical detector
        if self.detector.shape == "spherical":
            r = np.matrix([[np.cos(theta_i) * np.cos(phi_i), -np.cos(theta) * np.cos(phi)],
                           [np.sin(theta_i) * np.cos(phi_i), -np.sin(theta) * np.cos(phi)],
                           [np.sin(phi_i),                   -np.sin(phi)]])

            r_tt = np.matrix([[0,  np.sin(theta) * np.cos(phi)],
                              [0, -np.cos(theta) * np.cos(phi)],
                              [0,  0]])


            r_ph = np.matrix([[0,  np.cos(theta) * np.sin(phi)],
                              [0,  np.sin(theta) * np.sin(phi)],
                              [0, -np.cos(phi)]])

        # Appendix A.2 and A.3 Violini :: cylindrical detector
        elif self.detector.shape == "cylindrical":
            # Appendix A.3 Violini :: cylindrical detector with vertical axis
            if ~hasattr(self.detector, "orientation") or self.detector.orientation == "vertical":
                r = np.matrix([[np.cos(theta_i) * np.cos(phi_i), -np.cos(theta)],
                               [np.sin(theta_i) * np.cos(phi_i), -np.sin(theta)],
                               [np.sin(phi_i),                   -np.tan(phi)]])

                r_tt = np.matrix([[0,  np.sin(theta)],
                                  [0, -np.cos(theta)],
                                  [0,  0]])

                r_ph = np.matrix([[0,   0],
                                  [0,   0],
                                  [0, -(1 + np.tan(phi) ** 2)]])

            # Appendix A.2 Violini :: cylindrical detector with horizontal axis
            elif self.detector.orientation == "horizontal":
                # TODO: Complete horizontal cylindrical detector
                r = np.matrix([[np.cos(theta_i) * np.cos(phi_i), -np.tan(theta)],
                               [np.sin(theta_i) * np.cos(phi_i), -np.sin(phi)],
                               [np.sin(phi_i),                   -np.cos(phi)]])

                r_tt = np.matrix([])

                r_ph = np.matrix([])

                raise DetectorError("Horizontal cylindrical detector not yet supported")
            else:
                raise DetectorError("Unsupported cylindrical detector specified: {0}".format(self.detector.orientation))
        else:
            raise DetectorError("Detector shape \"{0}\" not supported".format(self.detector.shape))

        # Detector-shape-independent equations
        r_tt_i = np.matrix([[-np.sin(theta_i) * np.cos(phi_i), 0],
                            [ np.cos(theta_i) * np.cos(phi_i), 0],
                            [ 0,                               0]])

        r_ph_i = np.matrix([[-np.cos(theta_i) * np.sin(phi_i), 0],
                            [-np.sin(theta_i) * np.sin(phi_i), 0],
                            [ np.cos(phi_i),                   0]])

        # Eq 11  Violini :: cov(xi)
        sigma_sq = np.matrix(np.diag([sigma_t ** 2,
                                      sigma_t_md ** 2,
                                      sigma_l_pm ** 2,
                                      sigma_l_ms ** 2,
                                      sigma_l_sd ** 2,
                                      sigma_theta_i ** 2,
                                      sigma_phi_i ** 2,
                                      sigma_theta ** 2,
                                      sigma_phi ** 2]))

        # Eq 11 Violini :: Jacobi J
        q_derivs = np.array([r * np.matrix([[-m_n / hbar * vi / ti, m_n / hbar * vf / tf * l_ms / l_pm]]).T,
                             r * np.matrix([[0, -m_n / hbar * vf / tf]]).T,
                             r * np.matrix([[m_n / hbar / ti, -m_n / hbar * vf / tf * l_ms / (vi * l_pm)]]).T,
                             r * np.matrix([[0, m_n / hbar * vf / tf / vi]]).T,
                             r * np.matrix([[0, m_n / hbar / tf]]).T,
                             r_tt_i * vel.T,
                             r_ph_i * vel.T,
                             r_tt * vel.T,
                             r_ph * vel.T]).T * 1e-10

        # Eqs. 14-18 Violini :: Energy Derivatives
        e_derivs = np.matrix([[-m_n * (l_pm ** 2 / ti ** 3 + l_sd ** 2 / tf ** 3 * l_ms / l_pm),
                                m_n * (l_sd ** 2 / tf ** 3),
                                m_n * (l_pm / ti ** 2 + l_sd ** 2 /  tf ** 3 * ti /  l_pm ** 2 * l_ms),
                               -m_n * (l_sd ** 2 / tf ** 3 * ti / l_pm),
                               -m_n * (l_sd / tf ** 2)]]) / (1e-3 * e)

        # Eq 11 Violini :: Jacobi J
        jacobi = np.matrix(np.zeros((4, sigma_sq.shape[0])))
        jacobi[:3, :] = q_derivs
        jacobi[3, :e_derivs.size] = e_derivs

        # Eq 11 Violini :: cov(Q, hw)
        sigma_qe = jacobi * sigma_sq * jacobi.T

        # M = J^-1
        reso = np.linalg.inv(sigma_qe)

        # Transform from (ki, ki_perp, Qz) to (Q_perp, Q_para, Q_z)
        angle_ki_q = get_angle_ki_Q(ki, kf, self.sample.get_q(Q))

        rot_ki_q = np.matrix(np.eye(4))
        rot_ki_q[:2, :2] = np.matrix([[np.cos(angle_ki_q), -np.sin(angle_ki_q)],
                                      [np.sin(angle_ki_q),  np.cos(angle_ki_q)]])

        # Congruent Transformation T' M T
        res = rot_ki_q.T * reso * rot_ki_q

        print(res)
        # Eq 61 Violini :: resolution prefactor
        fwhm = 0
        X = np.append(Q, W)
        for i in range(4):
            for j in range(4):
                fwhm += res[i, j] * X[i] * X[j]

        r0 = 2 * np.log(2) / fwhm

        return r0, res


    def calc_resolution(self, hkle):
        r"""For a scattering vector (H,K,L) and  energy transfers W, given
        experimental conditions specified in the object, calculates the
        resolution matrix RMS and Resolution prefactor R0 in a coordinate
        system defined by the crystallographic axes of the sample,
        (h, k, l, w) in reciprocal lattice units (rlu).

        Parameters
        ----------
        hkle : list
            Array of the scattering vector and energy transfer at which the
            calculation should be performed

        Notes
        -----
            Translated from "A method to compute the covariance matrix of
            wavevector-energy transfer for neutron time-of-flight spectrometers"
            by Violini et al., Nuclear Instruments and Methods in Physics Research
            A 736 (2014) 31-39. DOI: 10.1016/j.nima.2013.10.042

        """
        self.HKLE = hkle
        [length, self.H, self.K, self.L, self.W] = _CleanArgs(*self.HKLE)

        r0_f = np.zeros(length)
        rm_f = np.zeros((length, 4, 4))
        rms_f = np.zeros((length, 4, 4))

        for i in range(length):
            print([self.H[i], self.K[i], self.L[i]], self.W[i])
            r0, rm = self.calc_resolution_in_Q_coords([self.H[i], self.K[i], self.L[i]], self.W[i])

            UBmat = np.eye(4)
            UBmat[:3, :3] = self.sample.UBmatrix

            vec = UBmat[:3, :3].dot(np.array([self.H[i], self.K[i], self.L[i]]))
            ang0 = -np.arctan2(vec[1], vec[0])

            Rot = np.matrix(np.eye(4))
            Rot[:2, :2] = np.matrix([[np.cos(-ang0), -np.sin(-ang0)],
                                     [np.sin(-ang0),  np.cos(-ang0)]])

            rms = (Rot * UBmat) * rm * (Rot * UBmat).T

            r0_f[i] = chop(r0)
            rm_f[i] = chop(rm)
            rms_f[i] = chop(rms)

        self.R0, self.RM, self.RMS = [np.squeeze(item) for item in [r0_f, rm_f, rms_f]]



r"""Instrument resolution calculations

"""

from .instrument import Instrument
from .monochromator import Monochromator
from .analyzer import Analyzer
from .chopper import Chopper
from .detector import Detector
from .general import GeneralInstrument
from .goniometer import Goniometer
from .guide import Guide
from .plot import PlotInstrument
from .tas_instrument import TripleAxisInstrument
from .tof_instrument import TimeOfFlightInstrument
from .tools import GetTau, get_angle_ki_Q, get_bragg_widths, get_kfree, chop
from . import tools
from . import exceptions



# -*- coding: utf-8 -*-
import numpy as np

from .tools import GetTau


class Monochromator(object):
    u"""Class containing monochromator information.

    Parameters
    ----------
    tau : float or str
        Tau value for the monochromator (or analyzer)

    mosaic : int
        Mosaic of the crystal in arc minutes

    direct : ±1, optional
        Direction of the crystal (left or right, -1 or +1, respectively).
        Default: -1 (left-handed coordinate frame).

    vmosaic : int, optional
        Vertical mosaic of the crystal in arc minutes. Default: None

    height : float, optional
        Height of the crystal in cm. Default: None

    width : float, optional
        Width of the crystal in cm. Default: None

    depth : float, optional
        Depth of the crystal in cm. Default: None

    Returns
    -------
    Monochromator : object

    Attributes
    ----------
    tau
    mosaic
    direct
    vmosaic
    height
    width
    depth
    rh
    rv
    d

    """

    def __init__(self, tau, mosaic, direct=-1, vmosaic=None, height=None, width=None, depth=None, rh=None, rv=None):
        self._tau = tau
        self.mosaic = mosaic
        if vmosaic is not None:
            self.vmosaic = vmosaic
        self.dir = direct
        self.d = 2 * np.pi / GetTau(tau)
        if rh is not None:
            self.rh = rh
        if rv is not None:
            self.rv = rv
        if height is not None:
            self.height = height
        if width is not None:
            self.width = width
        if depth is not None:
            self.depth = depth

    def __repr__(self):
        args = ', '.join([str(getattr(self, key)) for key in ['tau', 'mosaic']])
        kwargs = ', '.join(['{0}={1}'.format(key, getattr(self, key)) for key in
                            ['direct', 'vmosaic', 'height', 'width', 'depth', 'rh', 'rv'] if
                            getattr(self, key, None) is not None])
        return "Monochromator({0})".format(', '.join([args, kwargs]))

    def __eq__(self, right):
        self_parent_keys = sorted(list(self.__dict__.keys()))
        right_parent_keys = sorted(list(right.__dict__.keys()))

        if not np.all(self_parent_keys == right_parent_keys):
            return False

        for key, value in self.__dict__.items():
            right_parent_val = getattr(right, key)
            if not np.all(value == right_parent_val):
                return False

        return True

    def __ne__(self, right):
        return not self.__eq__(right)

    @property
    def tau(self):
        return self._tau

    @tau.setter
    def tau(self, tau):
        self._tau = tau
        self.d = 2 * np.pi / GetTau(tau)

    @property
    def direct(self):
        return self.dir

    @direct.setter
    def direct(self, value):
        self.dir = value



# -*- coding: utf-8 -*-
r"""Scattering methods for Materials class

"""


class HKLGenerator(object):
    r"""Methods for generating HKL reflections

    """
    def generate_hkl_positions(self):
        r"""Generates hkl positions where scattering is allowed
        """
        pass

    def find_equivalent_positions(self):
        r"""Eliminates equivalent hkl positions
        """
        pass

    def apply_scattering_rules(self):
        r"""Applies scattering rules for the space group
        """
        pass

    def find_site_multiplicity(self):
        r"""Finds the multiplicity of the hkl position
        """
        pass



# -*- coding: utf-8 -*-
import numbers

import numpy as np

from ..constants import BOLTZMANN_IN_MEV_K
from ..energy import Energy


class Analysis(object):
    r"""Class containing methods for the Data class

    Attributes
    ----------
    detailed_balance_factor

    Methods
    -------
    integrate
    position
    width
    scattering_function
    dynamic_susceptibility
    estimate_background
    get_keys
    get_bounds

    """

    @property
    def detailed_balance_factor(self):
        r"""Returns the detailed balance factor (sometimes called the Bose
        factor)

        Parameters
        ----------
            None

        Returns
        -------
        dbf : ndarray
            The detailed balance factor (temperature correction)

        """

        return 1. - np.exp(-self.Q[:, 3] / BOLTZMANN_IN_MEV_K / self.temp)

    def integrate(self, bounds=None, background=None, hkle=True):
        r"""Returns the integrated intensity within given bounds

        Parameters
        ----------
        bounds : bool, optional
            A boolean expression representing the bounds inside which the
            calculation will be performed

        background : float or dict, optional
            Default: None

        hkle : bool, optional
            If True, integrates only over h, k, l, e dimensions, otherwise
            integrates over all dimensions in :py:attr:`.Data.data`

        Returns
        -------
        result : float
            The integrated intensity either over all data, or within
            specified boundaries

        """
        result = 0
        for key in self.get_keys(hkle):
            result += np.trapz(self.intensity[self.get_bounds(bounds)] - self.estimate_background(background),
                               np.squeeze(self.data[key][self.get_bounds(bounds)]))

        return result

    def position(self, bounds=None, background=None, hkle=True):
        r"""Returns the position of a peak within the given bounds

        Parameters
        ----------
        bounds : bool, optional
            A boolean expression representing the bounds inside which the
            calculation will be performed

        background : float or dict, optional
            Default: None

        hkle : bool, optional
            If True, integrates only over h, k, l, e dimensions, otherwise
            integrates over all dimensions in :py:attr:`.Data.data`

        Returns
        -------
        result : tup
            The result is a tuple with position in each dimension of Q,
            (h, k, l, e)

        """
        result = ()
        for key in self.get_keys(hkle):
            _result = 0
            for key_integrate in self.get_keys(hkle):
                _result += np.trapz(self.data[key][self.get_bounds(bounds)] *
                                    (self.intensity[self.get_bounds(bounds)] - self.estimate_background(background)),
                                    self.data[key_integrate][self.get_bounds(bounds)]) / self.integrate(bounds, background)

            result += (np.squeeze(_result),)

        if hkle:
            return result
        else:
            return dict((key, value) for key, value in zip(self.get_keys(hkle), result))

    def width(self, bounds=None, background=None, fwhm=False, hkle=True):
        r"""Returns the mean-squared width of a peak within the given bounds

        Parameters
        ----------
        bounds : bool, optional
            A boolean expression representing the bounds inside which the
            calculation will be performed

        background : float or dict, optional
            Default: None

        fwhm : bool, optional
            If True, returns width in fwhm, otherwise in mean-squared width.
            Default: False

        hkle : bool, optional
            If True, integrates only over h, k, l, e dimensions, otherwise
            integrates over all dimensions in :py:attr:`.Data.data`

        Returns
        -------
        result : tup
            The result is a tuple with the width in each dimension of Q,
            (h, k, l, e)

        """

        result = ()
        for key in self.get_keys(hkle):
            _result = 0
            for key_integrate in self.get_keys(hkle):
                _result += np.trapz((self.data[key][self.get_bounds(bounds)] -
                                     self.position(bounds, background, hkle=False)[key]) ** 2 *
                                    (self.intensity[self.get_bounds(bounds)] - self.estimate_background(background)),
                                    self.data[key_integrate][self.get_bounds(bounds)]) / self.integrate(bounds, background)

            if fwhm:
                result += (np.sqrt(np.squeeze(_result)) * 2. * np.sqrt(2. * np.log(2.)),)
            else:
                result += (np.squeeze(_result),)

        if hkle:
            return result
        else:
            return dict((key, value) for key, value in zip(self.get_keys(hkle), result))

    def scattering_function(self, material, ei):
        r"""Returns the neutron scattering function, i.e. the detector counts
        scaled by :math:`4 \pi / \sigma_{\mathrm{tot}} * k_i/k_f`.

        Parameters
        ----------
        material : object
            Definition of the material given by the :py:class:`.Material`
            class

        ei : float
            Incident energy in meV

        Returns
        -------
        counts : ndarray
            The detector counts scaled by the total scattering cross section
            and ki/kf
        """
        ki = Energy(energy=ei).wavevector
        kf = Energy(energy=ei - self.e).wavevector

        return 4 * np.pi / material.total_scattering_cross_section * ki / kf * self.detector

    def dynamic_susceptibility(self, material, ei):
        r"""Returns the dynamic susceptibility
        :math:`\chi^{\prime\prime}(\mathbf{Q},\hbar\omega)`

        Parameters
        ----------
        material : object
            Definition of the material given by the :py:class:`.Material`
            class

        ei : float
            Incident energy in meV

        Returns
        -------
        counts : ndarray
            The detector counts turned into the scattering function multiplied
            by the detailed balance factor
        """
        return self.scattering_function(material, ei) * self.detailed_balance_factor

    def estimate_background(self, bg_params):
        r"""Estimate the background according to ``type`` specified.

        Parameters
        ----------
        bg_params : dict
            Input dictionary has keys 'type' and 'value'. Types are
                * 'constant' : background is the constant given by 'value'
                * 'percent' : background is estimated by the bottom x%, where x
                  is value
                * 'minimum' : background is estimated as the detector counts

        Returns
        -------
        background : float or ndarray
            Value determined to be the background. Will return ndarray only if
            `'type'` is `'constant'` and `'value'` is an ndarray
        """
        if isinstance(bg_params, type(None)):
            return 0

        elif isinstance(bg_params, numbers.Number):
            return bg_params

        elif bg_params['type'] == 'constant':
            return bg_params['value']

        elif bg_params['type'] == 'percent':
            inten = self.intensity[self.intensity >= 0.]
            Npts = int(inten.size * (bg_params['value'] / 100.))
            min_vals = inten[np.argsort(inten)[:Npts]]
            background = np.average(min_vals)
            return background

        elif bg_params['type'] == 'minimum':
            return min(self.intensity)

        else:
            return 0

    def get_bounds(self, bounds):
        r"""Generates a to_fit tuple if bounds is present in kwargs

        Parameters
        ----------
        bounds : dict

        Returns
        -------
        to_fit : tuple
            Tuple of indices

        """
        if bounds is not None:
            return np.where(bounds)
        else:
            return np.where(self.Q[:, 0])

    def get_keys(self, hkle):
        r"""Returns all of the Dictionary key names

        Parameters
        ----------
        hkle : bool
            If True only returns keys for h,k,l,e, otherwise returns all keys

        Returns
        -------
        keys : list
            :py:attr:`.Data.data` dictionary keys

        """
        if hkle:
            return [key for key in self.data if key in self.Q_keys.values()]
        else:
            return [key for key in self.data if key not in self.data_keys.values()]



import numpy as np

from ..lsfit import Fitter
from ..lsfit.tools import convert_params


class PlotData(object):
    """Class containing data plotting methods

    Methods
    -------
    plot
    plot_line
    plot_contour
    plot_volume

    """

    def plot(self, x=None, y=None, z=None, w=None, show_err=True, to_bin=None,
             plot_options=None, fit_options=None, smooth_options=None,
             output_file='', show_plot=True, **kwargs):
        r"""Plots the data in the class. x and y must at least be specified,
        and z and/or w being specified will produce higher dimensional plots
        (contour and volume, respectively).

        Parameters
        ----------
        x : str, optional
            `data_column` key defining the x-axis.
            Default: :py:attr:`plot_default_x`.

        y : str, optional
            `data_column` key defining the y-axis.
            Default: :py:attr:`plot_default_y`.

        z : str, optional
            `data_column` key defining the z-axis.
            Default: None

        w : str, optional
            `data_column` key defining the w-axis.
            Default: None

        bounds : dict, optional
            If set, data will be rebinned to the specified parameters, in the
            format `[min, max, num points]` for each `data_column` key. See
            documentation for :py:meth:`.Data.bin`. Default: None

        show_err : bool, optional
            Plot error bars. Only applies to xy scatter plots. Default: True

        show_plot : bool, optional
            Execute `plt.show()` to show the plot. Incompatible with
            `output_file` param. Default: True

        output_file : str, optional
            If set, the plot will be saved to the location given, in the format
            specified, provided that the format is supported. Default: None

        plot_options : dict, optional
            Plot options to be passed to the the matplotlib plotting routine.
            Default: None

        fit_options : dict, optional
            Fitting options to be passed to the Fitter routine. Default: None

        smooth_otions : dict, optional
            Smoothing options for Gaussian smoothing from
            `scipy.ndimage.filters.gaussian_filter`. Default: None

        kwargs : optional
            Additional plotting keyword arguments passed to the plotting
            function.

        """
        if to_bin is None:
            to_bin = dict()
        if plot_options is None:
            plot_options = dict()
        if fit_options is None:
            fit_options = dict()
        if smooth_options is None:
            smooth_options = dict(sigma=0)

        if x is None:
            try:
                x = self.plot_default_x
            except AttributeError:
                raise

        if y is None:
            try:
                y = self.plot_default_y
            except AttributeError:
                raise

        if w is not None:
            self.plot_volume(x, y, z, w, to_bin, plot_options, smooth_options,
                             output_file, show_plot, **kwargs)

        elif w is None and z is not None:
            self.plot_contour(x, y, z, to_bin, plot_options, smooth_options,
                              output_file, show_plot, **kwargs)

        elif w is None and z is None:
            self.plot_line(x, y, show_err, to_bin, plot_options, fit_options,
                           smooth_options, output_file, show_plot, **kwargs)

    def plot_volume(self, x, y, z, w, to_bin=None, plot_options=None, smooth_options=None, output_file='', show_plot=True, **kwargs):
        r"""Plots a 3D volume of 4D data

        Parameters
        ----------
        x : str
            `data_column` key defining the x-axis.
            Default: :py:attr:`plot_default_x`.

        y : str
            `data_column` key defining the y-axis.
            Default: :py:attr:`plot_default_y`.

        z : str
            `data_column` key defining the z-axis.
            Default: None

        w : str
            `data_column` key defining the w-axis.
            Default: None

        bounds : dict, optional
            If set, data will be rebinned to the specified parameters, in the
            format `[min, max, num points]` for each `data_column` key. See
            documentation for :py:meth:`.Data.bin`. Default: None

        show_err : bool, optional
            Plot error bars. Only applies to xy scatter plots. Default: True

        show_plot : bool, optional
            Execute `plt.show()` to show the plot. Incompatible with
            `output_file` param. Default: True

        output_file : str, optional
            If set, the plot will be saved to the location given, in the format
            specified, provided that the format is supported. Default: None

        plot_options : dict, optional
            Plot options to be passed to the the matplotlib plotting routine.
            Default: None

        fit_options : dict, optional
            Fitting options to be passed to the Fitter routine. Default: None

        smooth_otions : dict, optional
            Smoothing options for Gaussian smoothing from
            `scipy.ndimage.filters.gaussian_filter`. Default: None

        kwargs : optional
            Additional plotting keyword arguments passed to the plotting
            function.
        """
        try:
            import matplotlib.pyplot as plt
            from matplotlib import colors
            from mpl_toolkits.mplot3d import Axes3D
        except ImportError:
            raise ImportError('Matplotlib >= 1.3.0 is necessary for plotting.')

        if to_bin:
            data_bin = self.bin(to_bin)
            _x = data_bin.data[x]
            _y = data_bin.data[y]
            _z = data_bin.data[z]
            if w == 'intensity':
                _w = data_bin.intensity
            else:
                _w = data_bin.data[w]
        else:
            _x = self.data[x]
            _y = self.data[y]
            _z = self.data[z]
            if w == 'intensity':
                _w = self.intensity
            else:
                _w = self.data[w]

        if smooth_options['sigma'] > 0:
            from scipy.ndimage.filters import gaussian_filter
            _w = gaussian_filter(_w, **smooth_options)

        _x, _y, _z, _w = (np.ma.masked_where(_w <= 0, _x),
                          np.ma.masked_where(_w <= 0, _y),
                          np.ma.masked_where(_w <= 0, _z),
                          np.ma.masked_where(_w <= 0, _w))

        fig = plt.figure()
        axis = fig.add_subplot(111, projection='3d')

        axis.scatter(_x, _y, _z, c=_w, linewidths=0, vmin=1.e-4,
                     vmax=0.1, norm=colors.LogNorm())

        if output_file:
            plt.savefig(output_file)
        elif show_plot:
            plt.show()
        else:
            pass

    def plot_contour(self, x, y, z, to_bin=None, plot_options=None, smooth_options=None, output_file='', show_plot=True, **kwargs):
        r"""Method for plotting a 2D contour plot of 3D data

        Parameters
        ----------
        x : str
            `data_column` key defining the x-axis.
            Default: :py:attr:`plot_default_x`.

        y : str
            `data_column` key defining the y-axis.
            Default: :py:attr:`plot_default_y`.

        z : str
            `data_column` key defining the z-axis.
            Default: None

        bounds : dict, optional
            If set, data will be rebinned to the specified parameters, in the
            format `[min, max, num points]` for each `data_column` key. See
            documentation for :py:meth:`.Data.bin`. Default: None

        show_err : bool, optional
            Plot error bars. Only applies to xy scatter plots. Default: True

        show_plot : bool, optional
            Execute `plt.show()` to show the plot. Incompatible with
            `output_file` param. Default: True

        output_file : str, optional
            If set, the plot will be saved to the location given, in the format
            specified, provided that the format is supported. Default: None

        plot_options : dict, optional
            Plot options to be passed to the the matplotlib plotting routine.
            Default: None

        fit_options : dict, optional
            Fitting options to be passed to the Fitter routine. Default: None

        smooth_otions : dict, optional
            Smoothing options for Gaussian smoothing from
            `scipy.ndimage.filters.gaussian_filter`. Default: None

        kwargs : optional
            Additional plotting keyword arguments passed to the plotting
            function.
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise

        if to_bin:
            data_bin = self.bin(to_bin)
            _x = data_bin.data[x]
            _y = data_bin.data[y]
            if z == 'intensity':
                _z = data_bin.intensity
            else:
                _z = data_bin.data[z]
        else:
            _x = self.data[x]
            _y = self.data[y]
            if z == 'intensity':
                _z = self.intensity
            else:
                _z = self.data[z]

        if smooth_options['sigma'] > 0:
            from scipy.ndimage.filters import gaussian_filter
            _z = gaussian_filter(_z, **smooth_options)

        x_step = np.around(
            np.abs(np.unique(_x) - np.roll(np.unique(_x), 1))[1], decimals=4)
        y_step = np.around(
            np.abs(np.unique(_y) - np.roll(np.unique(_y), 1))[1], decimals=4)
        x_sparse = np.linspace(
            _x.min(), _x.max(), (_x.max() - _x.min()) / x_step + 1)
        y_sparse = np.linspace(
            _y.min(), _y.max(), (_y.max() - _y.min()) / y_step + 1)
        X, Y = np.meshgrid(x_sparse, y_sparse)

        from scipy.interpolate import griddata

        Z = griddata((_x, _y), _z, (X, Y))

        plt.pcolormesh(X, Y, Z, **plot_options)

        if output_file:
            plt.savefig(output_file)
        elif show_plot:
            plt.show()
        else:
            pass

    def plot_line(self, x, y, show_err=True, to_bin=None, plot_options=None, fit_options=None, smooth_options=None, output_file='', show_plot=True, **kwargs):
        r"""Method to Plot a line of 2D data

        Parameters
        ----------
        x : str
            `data_column` key defining the x-axis.
            Default: :py:attr:`plot_default_x`.

        y : str
            `data_column` key defining the y-axis.
            Default: :py:attr:`plot_default_y`.

        bounds : dict, optional
            If set, data will be rebinned to the specified parameters, in the
            format `[min, max, num points]` for each `data_column` key. See
            documentation for :py:meth:`.Data.bin`. Default: None

        show_err : bool, optional
            Plot error bars. Only applies to xy scatter plots. Default: True

        show_plot : bool, optional
            Execute `plt.show()` to show the plot. Incompatible with
            `output_file` param. Default: True

        output_file : str, optional
            If set, the plot will be saved to the location given, in the format
            specified, provided that the format is supported. Default: None

        plot_options : dict, optional
            Plot options to be passed to the the matplotlib plotting routine.
            Default: None

        fit_options : dict, optional
            Fitting options to be passed to the Fitter routine. Default: None

        smooth_otions : dict, optional
            Smoothing options for Gaussian smoothing from
            `scipy.ndimage.filters.gaussian_filter`. Default: None

        kwargs : optional
            Additional plotting keyword arguments passed to the plotting
            function.

        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise

        if to_bin:
            data_bin = self.bin(to_bin)
            _x = data_bin.data[x]
            if y == 'intensity':
                _y = data_bin.intensity
                _err = data_bin.error
            else:
                _y = data_bin.data[y]
                _err = np.sqrt(data_bin.data[y])
        else:
            _x = self.data[x]
            if y == 'intensity':
                _y = self.intensity
                _err = self.error
            else:
                _y = self.data[y]
                _err = np.sqrt(self.data[y])

        if smooth_options['sigma'] > 0:
            from scipy.ndimage.filters import gaussian_filter
            _y = gaussian_filter(_y, **smooth_options)

        if not plot_options:
            plot_options['fmt'] = 'rs'

        if show_err:
            plt.errorbar(_x, _y, yerr=_err, **plot_options)
        else:
            plt.errorbar(_x, _y, **plot_options)

        # add axis labels
        plt.xlabel(x)
        plt.ylabel(y)

        if fit_options:
            def residuals(params, data):
                funct, x, y, err = data

                return (y - funct(params, x)) / err

            fitobj = Fitter(residuals, data=(
                fit_options['function'], _x, _y, _err))
            if 'fixp' in fit_options:
                fitobj.parinfo = [{'fixed': fix}
                                  for fix in fit_options['fixp']]
            try:
                fitobj.fit(params0=fit_options['p'])
                fit_x = np.linspace(min(_x), max(_x), len(_x) * 10)
                fit_y = fit_options['function'](fitobj.params, fit_x)
                plt.plot(fit_x, fit_y, '{0}-'.format(plot_options['fmt'][0]))

                param_string = u'\n'.join(['p$_{{{0:d}}}$: {1:.3f}'.format(i, p)
                                           for i, p in enumerate(fitobj.params)])
                chi2_params = u'$\chi^2$: {0:.3f}\n\n'.format(
                    fitobj.chi2_min) + param_string

                plt.annotate(chi2_params, xy=(0.05, 0.95), xycoords='axes fraction',
                             horizontalalignment='left', verticalalignment='top',
                             bbox=dict(alpha=0.75, facecolor='white', edgecolor='none'))

            except Exception as mes:  # pylint: disable=broad-except
                raise Exception("Something wrong with fit: {0}".format(mes))

        if output_file:
            plt.savefig(output_file)
        elif show_plot:
            plt.show()
        else:
            pass



# -*- coding: utf-8 -*-
import numpy as np

from ..constants import periodic_table, scattering_lengths


class Atom(object):
    r"""Class for adding atoms to the Material class.

    Parameters
    ----------
    ion : string
        The name of the Atom, or ion if necessary

    pos : list(3)
        The position of the Atom in the chosen geometry

    dpos : list(3), optional
        Deviations from the position pos

    occupancy: float, optional
        Occupancy of the _Atom (*e.g.* if there is partial occupancy from
        doping)

    Mcell : float, optional
        The mass of the unit cell. If assigned, normalize scattering lengths to
        the square-root of the mass of the atom

    Returns
    -------
    output : object
        Atom object defining an individual atom in a unit cell of a single
        crystal

    """

    def __init__(self, ion, pos, occupancy=1., Mcell=None, massNorm=False, Uiso=0, Uaniso=np.zeros((3, 3))):
        self.ion = ion
        self.pos = np.array(pos)
        self.occupancy = occupancy
        self.Mcell = Mcell
        self.Uiso = Uiso
        self.Uaniso = np.matrix(Uaniso)

        if isinstance(scattering_lengths()[ion]['Coh b'], list):
            b = complex(*scattering_lengths()[ion]['Coh b'])
        else:
            b = scattering_lengths()[ion]['Coh b']

        if massNorm is True:
            self.mass = periodic_table()[ion]['mass']

            self.b = (b * self.occupancy * self.Mcell / np.sqrt(self.mass))
        else:
            self.b = b / 10.

        self.coh_xs = scattering_lengths()[ion]['Coh xs']
        self.inc_xs = scattering_lengths()[ion]['Inc xs']
        self.abs_xs = scattering_lengths()[ion]['Abs xs']

    def __repr__(self):
        return "Atom('{0}')".format(self.ion)


class MagneticAtom(object):
    r"""Class for adding magnetic atoms to the Material class.

    Parameters
    ----------
    ion : str
        The name of the ion

    pos : list(3)
        The position of the atom in r.l.u.

    Return
    ------
    output : object
        MagneticAtom object defining an individual magnetic ion in a unit cell

    """

    def __init__(self, ion, pos, moment, occupancy):
        self.ion = ion
        self.pos = np.array(pos)
        self.moment = moment
        self.occupancy = occupancy

    def __repr__(self):
        return "MagneticAtom('{0}')".format(self.ion, self.pos, self.moment, self.occupancy)
